#include "../std.h"
#include "databuffer.h"

DataBuffer::DataBuffer(unsigned int size_) {
  start=0; end=0; buf=NULL; read_start=0; read_length=0;
  subsize=size_/3; size=subsize*3;
  pthread_mutex_init(&lock,NULL);
  if(size > 0) buf=(char*)malloc(size);
}

DataBuffer::~DataBuffer(void) {
  if(buf) free(buf);
  pthread_mutex_destroy(&lock);
}

bool DataBuffer::for_read(char** start_,unsigned int *length_) {
  pthread_mutex_lock(&lock);
  *length_=0;
  if(!buf) { pthread_mutex_unlock(&lock); return false; };
  if(start == end) { start=0; end=0; };
  read_start = end;
  if(read_start >= size) read_start=0;
  read_length = size-read_start;
  if(start >= read_start) read_length=start-read_start-1;
  if(read_length > 2*subsize) read_length=2*subsize;
  if(read_length < 0) read_length=0;
  *length_=read_length;
  *start_=buf+read_start;
  pthread_mutex_unlock(&lock);
  return true;
}

bool DataBuffer::is_read(unsigned int length_) {
  pthread_mutex_lock(&lock);
  if(!buf) { pthread_mutex_unlock(&lock); return false; };
  bool res = true;
  if(length_ > read_length) { length_=read_length; res=false; };
  end=read_start+length_; read_start=end; read_length=0;
  pthread_mutex_unlock(&lock);
  return res;
}

bool DataBuffer::for_write(char** start_,unsigned int *length_) {
  pthread_mutex_lock(&lock);
  *length_=0;
  if(!buf) { pthread_mutex_unlock(&lock); return false; };
  long long int write_start = start;
  long long int write_length = end - start;
  if(end < start) write_length=size-start;
  if(write_length >= subsize*2) write_length=subsize*2;
  *length_=write_length;
  *start_=buf+write_start;
  pthread_mutex_unlock(&lock);
  return true;
}

bool DataBuffer::is_written(unsigned int length_) {
  pthread_mutex_lock(&lock);
  if(!buf) { pthread_mutex_unlock(&lock); return false; };
  bool res = true;
  if(start <= end) {
    start+=length_;
    if(start > end) {
      res=false; start=0; end=0;
    };
  }
  else {
    start+=length_;
    if(start > size) res=false;
    if(start >= size) start=0;
  };
  pthread_mutex_unlock(&lock);
  return res;
}

