#include "../std.h"
#include <iostream>

#include "../misc/url_options.h"
#include "../misc/stringtoint.h"
#include "../misc/mkdir_recursive.h"
#include "../misc/checkfile.h"
#include "../misc/log_time.h"
#include "datapoint.h"
#include "databufferpar.h"
#include "datahandle.h"
#include "datacallback.h"

void* DataHandle::read_file(void* arg) {
  DataHandle* it = (DataHandle*)arg;
  bool limit_length = false;
  unsigned long long int range_length;
  unsigned long long int offset = 0;
  if(it->range_end > it->range_start) {
    range_length = it->range_end - it->range_start;
    limit_length = true;
    lseek64(it->fd,it->range_start,SEEK_SET);
    offset=it->range_start;
  } else {
    lseek64(it->fd,0,SEEK_SET);
  };
  for(;;) {
    if(limit_length) if(range_length==0) break;
  /* read from fd here and push to buffer */
  /* 1. claim buffer */
    int h;
    unsigned int l;
    if(!it->buffer->for_read(h,l,true)) { 
      /* failed to get buffer - must be error or request to exit */
      it->buffer->error_read(true);
      break;
    };
    if(it->buffer->error()) {
      it->buffer->is_read(h,0,0);
      break;
    };
  /* 2. read */
    if(limit_length) if(l>range_length) l=range_length;
    unsigned long long int p = lseek64(it->fd,0,SEEK_CUR);
    if(p == (unsigned long long int)(-1)) p=offset;
    int ll = read(it->fd,(*(it->buffer))[h],l);
    if(ll == -1) { /* error */
      it->buffer->is_read(h,0,0);
      it->buffer->error_read(true);
      break;
    };
    if(ll == 0) { /* eof */
      it->buffer->is_read(h,0,0);
      break;
    };
  /* 3. announce */
    it->buffer->is_read(h,ll,p);
    if(limit_length) {
      if(ll>range_length) { range_length=0; } else { range_length-=ll; };
    };
    offset+=ll; // for non-seakable files
  };
  close(it->fd);
  it->buffer->eof_read(true);
  /* inform parent thread about exit */
  it->file_thread_exited.signal();
  return NULL;
}

void* DataHandle::write_file(void* arg) {
  DataHandle* it = (DataHandle*)arg;
  for(;;) {
  /* take from buffer and write to fd */
  /* 1. claim buffer */
    int h;
    unsigned int l;
    unsigned long long int p;
    if(!it->buffer->for_write(h,l,p,true)) {
      /* failed to get buffer - must be error or request to exit */
      if(!it->buffer->eof_read()) it->buffer->error_write(true);
      it->buffer->eof_write(true);
      break;
    };
    if(it->buffer->error()) {
      it->buffer->is_written(h);
      it->buffer->eof_write(true);
      break;
    };
  /* 2. write */
    lseek64(it->fd,p,SEEK_SET);
    int l_=0;
    int ll = 0;
    for(;l_<l;) {
      ll = write(it->fd,(*(it->buffer))[h]+l_,l-l_);
      if(ll == -1) { /* error */
        it->buffer->is_written(h);
        it->buffer->error_write(true);
        it->buffer->eof_write(true);
        break;
      };
      l_+=ll;
    };
    if(ll == -1) break;
  /* 3. announce */
    it->buffer->is_written(h);
  };
  close(it->fd);
  /* inform parent thread about exit */
  it->file_thread_exited.signal();
  return NULL;
}

bool DataHandle::check_file(void) {
  const char* url_path = get_url_path(c_url.c_str());
  int res=check_file_access(url_path,O_RDONLY,get_user_id(),(gid_t)(-1));
  if(res != 0) {
    odlog(1)<<"File is not accessible: "<<url_path<<endl;
    return false;
  };
  struct stat64 st;
  if(stat64(url_path,&st) != 0) {
    odlog(1)<<"Can't stat file: "<<url_path<<endl;
    return false;
  };
  url->meta_size(st.st_size);
  url->meta_created(st.st_mtime);
  return true;
}

bool DataHandle::remove_file(void) {
  const char* path = get_url_path(c_url.c_str());
  if(unlink(path) == -1) {
    if(errno != ENOENT) return false;
  };
  return true;
}

bool DataHandle::start_reading_file(DataBufferPar &buf) {
  /* try to open */
  file_thread_exited.reset();
  if(!strcmp("-",c_url.c_str())) {
    fd=dup(STDIN_FILENO);
  } else {
    if(check_file_access(get_url_path(c_url.c_str()),O_RDONLY,
                       get_user_id(),(gid_t)(-1)) != 0) return false;
    fd=open64(get_url_path(c_url.c_str()),O_RDONLY);
  };
  if(fd == -1) return false;
  /* provide some metadata */
  struct stat64 st;
  if(fstat64(fd,&st) == 0) {
    url->meta_size(st.st_size);
    url->meta_created(st.st_mtime);
  };
  buffer=&buf;
  /* create thread to maintain reading */
  pthread_attr_init(&file_thread_attr);
  pthread_attr_setdetachstate(&file_thread_attr,PTHREAD_CREATE_DETACHED);
  if(pthread_create(&file_thread,&file_thread_attr,&read_file,this) != 0) {
    pthread_attr_destroy(&file_thread_attr);
    close(fd); fd=-1; return false;
  };
  return true;
}

bool DataHandle::stop_reading_file(void) {
  if(!buffer->eof_read()) {
    buffer->error_read(true);        /* trigger transfer error */
    close(fd); fd=-1;
  };
  file_thread_exited.wait();  /* wait till reading thread exited */
  pthread_attr_destroy(&file_thread_attr);
  return true;
}

bool DataHandle::start_writing_file(DataBufferPar &buf,DataCallback *space_cb) {
  /* try to open */
  file_thread_exited.reset();
  buffer=&buf;
  if(!strcmp("-",c_url.c_str())) {
    fd=dup(STDOUT_FILENO);
    if(fd == -1) {
      odlog(0)<<"Failed to use channel stdout"<<endl;
      buffer->error_write(true); buffer->eof_write(true);
      return false;
    };
  } else {
    uid_t uid=get_user_id();
    gid_t gid=get_user_group(uid);
    /* do not check permissions to create anything here - 
       suppose it path was checked at higher level */
    /* make directories */
    const char* path_ = get_url_path(c_url.c_str());
    if(path_ == NULL) {
      odlog(0)<<"Invalid url: "<<c_url<<endl;
      buffer->error_write(true); buffer->eof_write(true);
      return false;
    };
    string path = path_;
    {
      if(path[0] != '/') path="/"+path;
      string dirpath = path;
      int n = dirpath.rfind('/');
      if(n == 0) {
        dirpath="/";
      } else {
        dirpath.erase(n,dirpath.length()-n+1);
      };
      if(mkdir_recursive(NULL,dirpath.c_str(),S_IRWXU,uid,gid) != 0) {
        if(errno != EEXIST) {
          odlog(0)<<"Failed to create/find directory "<<dirpath<<endl;
          buffer->error_write(true); buffer->eof_write(true);
          return false;
        };
      }; 
    };
    /* try to create file, if failed - try to open it */
    fd=open64(path.c_str(),O_WRONLY | O_CREAT | O_EXCL,S_IRUSR | S_IWUSR);
    if(fd == -1) {
      fd=open64(path.c_str(),O_WRONLY | O_CREAT,S_IRUSR | S_IWUSR);
    }
    else { /* this file was created by us. Hence we can set it's owner */
      fchown(fd,uid,gid);
    };
    if(fd == -1) {
      odlog(0)<<"Failed to create/open file "<<path<<endl;
      buffer->error_write(true); buffer->eof_write(true);
      return false;
    };

    /* preallocate space */
    buffer->speed.hold(true);
    unsigned long long int fsize = url->meta_size();
    if(fsize > 0) {
      odlog(1)<<"setting file "<<get_url_path(c_url.c_str())<<
                " to size "<<fsize<<endl;
      /* because filesytem can skip empty blocks do real write */
      unsigned long long int old_size=lseek64(fd,0,SEEK_END);
      if(old_size < fsize) {
        char buf[65536];
        memset(buf,0xFF,sizeof(buf));
        unsigned int l=1;
        for(;l>0;) {
          old_size=lseek64(fd,0,SEEK_END);
          l=sizeof(buf);
          if(l > (fsize-old_size)) l=fsize-old_size;
          if(write(fd,buf,l) == -1) {
            /* out of space */
            perror("write"); /* not good, but let it be */
            if(space_cb != NULL) {
              if(space_cb->cb((unsigned long long int)l)) continue;
            };
            lseek64(fd,0,SEEK_SET); 
            ftruncate64(fd,0);
            close(fd); fd=-1; 
            odlog(1)<<"Failed to preallocate space"<<endl;
            buffer->speed.reset();
            buffer->speed.hold(false);
            buffer->error_write(true); buffer->eof_write(true);
            return false;
          };
        };
      };
    };
  };
  buffer->speed.reset();
  buffer->speed.hold(false);
  /* create thread to maintain writing */
  pthread_attr_init(&file_thread_attr);
  pthread_attr_setdetachstate(&file_thread_attr,PTHREAD_CREATE_DETACHED);
  if(pthread_create(&file_thread,&file_thread_attr,&write_file,this) != 0) {
    pthread_attr_destroy(&file_thread_attr);
    close(fd); fd=-1;
    buffer->error_write(true); buffer->eof_write(true);
    return false;
  };
  return true;
}

bool DataHandle::stop_writing_file(void) {
  if(!buffer->eof_write()) {
    buffer->error_write(true);        /* trigger transfer error */
    close(fd); fd=-1;
  };
  file_thread_exited.wait();  /* wait till reading thread exited */
  pthread_attr_destroy(&file_thread_attr);
  return true;
}

bool DataHandle::list_files_file(list<DataPoint::FileInfo> &files,bool resolve) {
  string dirname = get_url_path(c_url.c_str());
  if(dirname[dirname.length()-1] == '/') dirname.resize(dirname.length()-1);
  DIR *dir=opendir(dirname.c_str());
  if(dir == NULL) {
    odlog(1)<<"Failed to read directory: "<<dirname<<endl;
    return false;
  };
  for(;;) {
    struct dirent file_;
    struct dirent *file;
    readdir_r(dir,&file_,&file);
    if(file == NULL) break;
    if(file->d_name[0] == '.') {
      if(file->d_name[1] == 0) continue;
      if(file->d_name[1] == '.') if(file->d_name[2] == 0) continue;
    };
    list<DataPoint::FileInfo>::iterator f = 
               files.insert(files.end(),DataPoint::FileInfo(file->d_name));
    if(resolve) {
      string fname =  dirname;
      dirname+="/"; dirname+=file->d_name;
      struct stat64 st;
      if(stat64(fname.c_str(),&st) == 0) {
        f->size=st.st_size; f->size_available=true;
        f->created=st.st_mtime; f->created_available=true;
        if(S_ISDIR(st.st_mode)) {
          f->type=DataPoint::FileInfo::file_type_dir;
        } else if(S_ISREG(st.st_mode)) {
          f->type=DataPoint::FileInfo::file_type_dir;
        };
      };
    };
  }; 
  return true;
}

