#include "../../std.h"

#include <string>

#include "../../misc/log_time.h"
#include "../../misc/escaped.h"
#include "../../misc/time_utils.h"


// HACK
extern string base_gsi_url;
extern string base_ssl_url;


// one more hack
#define WITH_NOGLOBAL
#define SOAP_FMAC3 static
#include "srm_soapC.cpp"
#include "srm_soapClient.cpp"
#include "srm_soapServer.cpp"

#define __SRM_INTERNAL_INCLUDE__
#include "srm.h"

extern SOAP_NMAC struct Namespace srm_soap_namespaces[];


#define SOAP_ENV__Header SE_ENV__Header
#define SOAP_ENV__Code   SE_ENV__Code
#define SOAP_ENV__Detail SE_ENV__Detail
#define SOAP_ENV__Fault  SE_ENV__Fault
#include "../se/file_soapStub.h"
#undef SE_ENV__Header
#undef SE_ENV__Code
#undef SE_ENV__Detail
#undef SE_ENV__Fault

#ifdef OLD_GSOAP
#error gSOAP older than 2.5.2 is not supported.
#else
void HTTP_SRM::soap_methods(void) {
  odlog(3)<<"soap_methods: tag: "<<sp.tag<<endl;
  if(!soap_match_tag(&sp,sp.tag,"srm1:srmReserveSpace")) soap_serve_srm1__srmReserveSpace(&sp);
  else if(!soap_match_tag(&sp,sp.tag,"srm1:srmReleaseSpace")) soap_serve_srm1__srmReleaseSpace(&sp);
  else if(!soap_match_tag(&sp,sp.tag,"srm1:srmUpdateSpace")) soap_serve_srm1__srmUpdateSpace(&sp);
  else if(!soap_match_tag(&sp,sp.tag,"srm1:srmCompactSpace")) soap_serve_srm1__srmCompactSpace(&sp);
  else if(!soap_match_tag(&sp,sp.tag,"srm1:srmGetSpaceMetaData")) soap_serve_srm1__srmGetSpaceMetaData(&sp);
  else if(!soap_match_tag(&sp,sp.tag,"srm1:srmChangeFileStorageType")) soap_serve_srm1__srmChangeFileStorageType(&sp);
  else if(!soap_match_tag(&sp,sp.tag,"srm1:srmGetSpaceToken")) soap_serve_srm1__srmGetSpaceToken(&sp);
  else if(!soap_match_tag(&sp,sp.tag,"srm1:srmSetPermission")) soap_serve_srm1__srmSetPermission(&sp);
  else if(!soap_match_tag(&sp,sp.tag,"srm1:srmReassignToUser")) soap_serve_srm1__srmReassignToUser(&sp);
  else if(!soap_match_tag(&sp,sp.tag,"srm1:srmCheckPermission")) soap_serve_srm1__srmCheckPermission(&sp);
  else if(!soap_match_tag(&sp,sp.tag,"srm1:srmMkdir")) soap_serve_srm1__srmMkdir(&sp);
  else if(!soap_match_tag(&sp,sp.tag,"srm1:srmRmdir")) soap_serve_srm1__srmRmdir(&sp);
  else if(!soap_match_tag(&sp,sp.tag,"srm1:srmRm")) soap_serve_srm1__srmRm(&sp);
  else if(!soap_match_tag(&sp,sp.tag,"srm1:srmLs")) soap_serve_srm1__srmLs(&sp);
  else if(!soap_match_tag(&sp,sp.tag,"srm1:srmMv")) soap_serve_srm1__srmMv(&sp);
  else if(!soap_match_tag(&sp,sp.tag,"srm1:srmPrepareToGet")) soap_serve_srm1__srmPrepareToGet(&sp);
  else if(!soap_match_tag(&sp,sp.tag,"srm1:srmPrepareToPut")) soap_serve_srm1__srmPrepareToPut(&sp);
  else if(!soap_match_tag(&sp,sp.tag,"srm1:srmCopy")) soap_serve_srm1__srmCopy(&sp);
  else if(!soap_match_tag(&sp,sp.tag,"srm1:srmRemoveFiles")) soap_serve_srm1__srmRemoveFiles(&sp);
  else if(!soap_match_tag(&sp,sp.tag,"srm1:srmReleaseFiles")) soap_serve_srm1__srmReleaseFiles(&sp);
  else if(!soap_match_tag(&sp,sp.tag,"srm1:srmPutDone")) soap_serve_srm1__srmPutDone(&sp);
  else if(!soap_match_tag(&sp,sp.tag,"srm1:srmAbortRequest")) soap_serve_srm1__srmAbortRequest(&sp);
  else if(!soap_match_tag(&sp,sp.tag,"srm1:srmAbortFiles")) soap_serve_srm1__srmAbortFiles(&sp);
  else if(!soap_match_tag(&sp,sp.tag,"srm1:srmSuspendRequest")) soap_serve_srm1__srmSuspendRequest(&sp);
  else if(!soap_match_tag(&sp,sp.tag,"srm1:srmResumeRequest")) soap_serve_srm1__srmResumeRequest(&sp);
  else if(!soap_match_tag(&sp,sp.tag,"srm1:srmStatusOfGetRequest")) soap_serve_srm1__srmStatusOfGetRequest(&sp);
  else if(!soap_match_tag(&sp,sp.tag,"srm1:srmStatusOfPutRequest")) soap_serve_srm1__srmStatusOfPutRequest(&sp);
  else if(!soap_match_tag(&sp,sp.tag,"srm1:srmStatusOfCopyRequest")) soap_serve_srm1__srmStatusOfCopyRequest(&sp);
  else if(!soap_match_tag(&sp,sp.tag,"srm1:srmGetRequestSummary")) soap_serve_srm1__srmGetRequestSummary(&sp);
  else if(!soap_match_tag(&sp,sp.tag,"srm1:srmExtendFileLifeTime")) soap_serve_srm1__srmExtendFileLifeTime(&sp);
  else if(!soap_match_tag(&sp,sp.tag,"srm1:srmGetRequestID")) soap_serve_srm1__srmGetRequestID(&sp);
  else sp.error = SOAP_NO_METHOD;
}
#endif

// Non-supported methods

int srm1__srmReserveSpace(struct soap* sp, srm1__srmReserveSpaceRequest *req, struct srm1__srmReserveSpaceResponse_ &resp) {
  resp.srmReserveSpaceResponse = soap_new_srm1__srmReserveSpaceResponse(sp,-1);
  resp.srmReserveSpaceResponse->returnStatus=soap_new_srm1__TReturnStatus(sp,-1);
  resp.srmReserveSpaceResponse->returnStatus->statusCode=SRM_USCORENOT_USCORESUPPORTED;
  return SOAP_OK;
}

int srm1__srmReleaseSpace(struct soap* sp, srm1__srmReleaseSpaceRequest *req, struct srm1__srmReleaseSpaceResponse_ &resp) {
  resp.srmReleaseSpaceResponse = soap_new_srm1__srmReleaseSpaceResponse(sp,-1);
  resp.srmReleaseSpaceResponse->returnStatus=soap_new_srm1__TReturnStatus(sp,-1);
  resp.srmReleaseSpaceResponse->returnStatus->statusCode=SRM_USCORENOT_USCORESUPPORTED;
  return SOAP_OK;
}


int srm1__srmUpdateSpace(struct soap* sp, srm1__srmUpdateSpaceRequest *req, struct srm1__srmUpdateSpaceResponse_ &resp) {
  resp.srmUpdateSpaceResponse = soap_new_srm1__srmUpdateSpaceResponse(sp,-1);
  resp.srmUpdateSpaceResponse->returnStatus=soap_new_srm1__TReturnStatus(sp,-1);
  resp.srmUpdateSpaceResponse->returnStatus->statusCode=SRM_USCORENOT_USCORESUPPORTED;
  return SOAP_OK;
}


int srm1__srmCompactSpace(struct soap* sp, srm1__srmCompactSpaceRequest *req, struct srm1__srmCompactSpaceResponse_ &resp) {
  resp.srmCompactSpaceResponse = soap_new_srm1__srmCompactSpaceResponse(sp,-1);
  resp.srmCompactSpaceResponse->returnStatus=soap_new_srm1__TReturnStatus(sp,-1);
  resp.srmCompactSpaceResponse->returnStatus->statusCode=SRM_USCORENOT_USCORESUPPORTED;
  return SOAP_OK;
}


int srm1__srmGetSpaceMetaData(struct soap* sp, srm1__srmGetSpaceMetaDataRequest *req, struct srm1__srmGetSpaceMetaDataResponse_ &resp) {
  resp.srmGetSpaceMetaDataResponse = soap_new_srm1__srmGetSpaceMetaDataResponse(sp,-1);
  resp.srmGetSpaceMetaDataResponse->returnStatus=soap_new_srm1__TReturnStatus(sp,-1);
  resp.srmGetSpaceMetaDataResponse->returnStatus->statusCode=SRM_USCORENOT_USCORESUPPORTED;
  return SOAP_OK;
}


int srm1__srmChangeFileStorageType(struct soap* sp, srm1__srmChangeFileStorageTypeRequest *req, struct srm1__srmChangeFileStorageTypeResponse_ &resp) {
  resp.srmChangeFileStorageTypeResponse = soap_new_srm1__srmChangeFileStorageTypeResponse(sp,-1);
  resp.srmChangeFileStorageTypeResponse->returnStatus=soap_new_srm1__TReturnStatus(sp,-1);
  resp.srmChangeFileStorageTypeResponse->returnStatus->statusCode=SRM_USCORENOT_USCORESUPPORTED;
  return SOAP_OK;
}


int srm1__srmGetSpaceToken(struct soap* sp, srm1__srmGetSpaceTokenRequest *req, struct srm1__srmGetSpaceTokenResponse_ &resp) {
  resp.srmGetSpaceTokenResponse = soap_new_srm1__srmGetSpaceTokenResponse(sp,-1);
  resp.srmGetSpaceTokenResponse->returnStatus=soap_new_srm1__TReturnStatus(sp,-1);
  resp.srmGetSpaceTokenResponse->returnStatus->statusCode=SRM_USCORENOT_USCORESUPPORTED;
  return SOAP_OK;
}


int srm1__srmSetPermission(struct soap* sp, srm1__srmSetPermissionRequest *req, struct srm1__srmSetPermissionResponse_ &resp) {
  resp.srmSetPermissionResponse=soap_new_srm1__srmSetPermissionResponse(sp,-1);
  resp.srmSetPermissionResponse->returnStatus=soap_new_srm1__TReturnStatus(sp,-1);
  resp.srmSetPermissionResponse->returnStatus->statusCode=SRM_USCORENOT_USCORESUPPORTED;
  return SOAP_OK;
}


int srm1__srmReassignToUser(struct soap* sp, srm1__srmReassignToUserRequest *req, struct srm1__srmReassignToUserResponse_ &resp) {
  resp.srmReassignToUserResponse = soap_new_srm1__srmReassignToUserResponse(sp,-1);
  resp.srmReassignToUserResponse->returnStatus=soap_new_srm1__TReturnStatus(sp,-1);
  resp.srmReassignToUserResponse->returnStatus->statusCode=SRM_USCORENOT_USCORESUPPORTED;
  return SOAP_OK;
}


int srm1__srmCheckPermission(struct soap* sp, srm1__srmCheckPermissionRequest *req, struct srm1__srmCheckPermissionResponse_ &resp) {
  resp.srmCheckPermissionResponse = soap_new_srm1__srmCheckPermissionResponse(sp,-1);
  resp.srmCheckPermissionResponse->returnStatus=soap_new_srm1__TReturnStatus(sp,-1);
  resp.srmCheckPermissionResponse->returnStatus->statusCode=SRM_USCORENOT_USCORESUPPORTED;
  return SOAP_OK;
}


int srm1__srmMkdir(struct soap* sp, srm1__srmMkdirRequest *req, struct srm1__srmMkdirResponse_ &resp) {
  resp.srmMkdirResponse = soap_new_srm1__srmMkdirResponse(sp,-1);
  resp.srmMkdirResponse->returnStatus=soap_new_srm1__TReturnStatus(sp,-1);
  resp.srmMkdirResponse->returnStatus->statusCode=SRM_USCORENOT_USCORESUPPORTED;
  return SOAP_OK;
}


int srm1__srmRmdir(struct soap* sp, srm1__srmRmdirRequest *req, struct srm1__srmRmdirResponse_ &resp) {
  resp.srmRmdirResponse = soap_new_srm1__srmRmdirResponse(sp,-1);
  resp.srmRmdirResponse->returnStatus=soap_new_srm1__TReturnStatus(sp,-1);
  resp.srmRmdirResponse->returnStatus->statusCode=SRM_USCORENOT_USCORESUPPORTED;
  return SOAP_OK;
}


int srm1__srmRm(struct soap* sp, srm1__srmRmRequest *req, struct srm1__srmRmResponse_ &resp) {
  resp.srmRmResponse = soap_new_srm1__srmRmResponse(sp,-1);
  resp.srmRmResponse->returnStatus=soap_new_srm1__TReturnStatus(sp,-1);
  resp.srmRmResponse->returnStatus->statusCode=SRM_USCORENOT_USCORESUPPORTED;
  return SOAP_OK;
}


int srm1__srmMv(struct soap* sp, srm1__srmMvRequest *req, struct srm1__srmMvResponse_ &resp) {
  resp.srmMvResponse = soap_new_srm1__srmMvResponse(sp,-1);
  resp.srmMvResponse->returnStatus=soap_new_srm1__TReturnStatus(sp,-1);
  resp.srmMvResponse->returnStatus->statusCode=SRM_USCORENOT_USCORESUPPORTED;
  return SOAP_OK;
}


int srm1__srmPrepareToPut(struct soap* sp, srm1__srmPrepareToPutRequest *req, struct srm1__srmPrepareToPutResponse_ &resp) {
  resp.srmPrepareToPutResponse = soap_new_srm1__srmPrepareToPutResponse(sp,-1);
  resp.srmPrepareToPutResponse->returnStatus=soap_new_srm1__TReturnStatus(sp,-1);
  resp.srmPrepareToPutResponse->returnStatus->statusCode=SRM_USCORENOT_USCORESUPPORTED;
  return SOAP_OK;
}


int srm1__srmPrepareToGet(struct soap* sp, srm1__srmPrepareToGetRequest *req, struct srm1__srmPrepareToGetResponse_ &resp) {
  resp.srmPrepareToGetResponse = soap_new_srm1__srmPrepareToGetResponse(sp,-1);
  resp.srmPrepareToGetResponse->returnStatus=soap_new_srm1__TReturnStatus(sp,-1);
  resp.srmPrepareToGetResponse->returnStatus->statusCode=SRM_USCORENOT_USCORESUPPORTED;
  return SOAP_OK;
}


int srm1__srmCopy(struct soap* sp, srm1__srmCopyRequest *req, struct srm1__srmCopyResponse_ &resp) {
  resp.srmCopyResponse = soap_new_srm1__srmCopyResponse(sp,-1);
  resp.srmCopyResponse->returnStatus=soap_new_srm1__TReturnStatus(sp,-1);
  resp.srmCopyResponse->returnStatus->statusCode=SRM_USCORENOT_USCORESUPPORTED;
  return SOAP_OK;
}


int srm1__srmRemoveFiles(struct soap* sp, srm1__srmRemoveFilesRequest *req, struct srm1__srmRemoveFilesResponse_ &resp) {
  resp.srmRemoveFilesResponse = soap_new_srm1__srmRemoveFilesResponse(sp,-1);
  resp.srmRemoveFilesResponse->returnStatus=soap_new_srm1__TReturnStatus(sp,-1);
  resp.srmRemoveFilesResponse->returnStatus->statusCode=SRM_USCORENOT_USCORESUPPORTED;
  return SOAP_OK;
}


int srm1__srmReleaseFiles(struct soap* sp, srm1__srmReleaseFilesRequest *req, struct srm1__srmReleaseFilesResponse_ &resp) {
  resp.srmReleaseFilesResponse = soap_new_srm1__srmReleaseFilesResponse(sp,-1);
  resp.srmReleaseFilesResponse->returnStatus=soap_new_srm1__TReturnStatus(sp,-1);
  resp.srmReleaseFilesResponse->returnStatus->statusCode=SRM_USCORENOT_USCORESUPPORTED;
  return SOAP_OK;
}


int srm1__srmPutDone(struct soap* sp, srm1__srmPutDoneRequest*, struct srm1__srmPutDoneResponse_ &resp) {
  resp.srmPutDoneResponse = soap_new_srm1__srmPutDoneResponse(sp,-1);
  resp.srmPutDoneResponse->returnStatus=soap_new_srm1__TReturnStatus(sp,-1);
  resp.srmPutDoneResponse->returnStatus->statusCode=SRM_USCORENOT_USCORESUPPORTED;
  return SOAP_OK;
}


int srm1__srmAbortRequest(struct soap* sp, srm1__srmAbortRequestRequest *req, struct srm1__srmAbortRequestResponse_ &resp) {
  resp.srmAbortRequestResponse = soap_new_srm1__srmAbortRequestResponse(sp,-1);
  resp.srmAbortRequestResponse->returnStatus=soap_new_srm1__TReturnStatus(sp,-1);
  resp.srmAbortRequestResponse->returnStatus->statusCode=SRM_USCORENOT_USCORESUPPORTED;
  return SOAP_OK;
}


int srm1__srmAbortFiles(struct soap* sp, srm1__srmAbortFilesRequest *req, struct srm1__srmAbortFilesResponse_ &resp) {
  resp.srmAbortFilesResponse = soap_new_srm1__srmAbortFilesResponse(sp,-1);
  resp.srmAbortFilesResponse->returnStatus=soap_new_srm1__TReturnStatus(sp,-1);
  resp.srmAbortFilesResponse->returnStatus->statusCode=SRM_USCORENOT_USCORESUPPORTED;
  return SOAP_OK;
}


int srm1__srmSuspendRequest(struct soap* sp, srm1__srmSuspendRequestRequest *req, struct srm1__srmSuspendRequestResponse_ &resp) {
  resp.srmSuspendRequestResponse = soap_new_srm1__srmSuspendRequestResponse(sp,-1);
  resp.srmSuspendRequestResponse->returnStatus=soap_new_srm1__TReturnStatus(sp,-1);
  resp.srmSuspendRequestResponse->returnStatus->statusCode=SRM_USCORENOT_USCORESUPPORTED;
  return SOAP_OK;
}


int srm1__srmResumeRequest(struct soap* sp, srm1__srmResumeRequestRequest *req, struct srm1__srmResumeRequestResponse_ &resp) {
  resp.srmResumeRequestResponse = soap_new_srm1__srmResumeRequestResponse(sp,-1);
  resp.srmResumeRequestResponse->returnStatus=soap_new_srm1__TReturnStatus(sp,-1);
  resp.srmResumeRequestResponse->returnStatus->statusCode=SRM_USCORENOT_USCORESUPPORTED;
  return SOAP_OK;
}


int srm1__srmStatusOfGetRequest(struct soap* sp, srm1__srmStatusOfGetRequestRequest *req, struct srm1__srmStatusOfGetRequestResponse_ &resp) {
  resp.srmStatusOfGetRequestResponse = soap_new_srm1__srmStatusOfGetRequestResponse(sp,-1);
  resp.srmStatusOfGetRequestResponse->returnStatus=soap_new_srm1__TReturnStatus(sp,-1);
  resp.srmStatusOfGetRequestResponse->returnStatus->statusCode=SRM_USCORENOT_USCORESUPPORTED;
  return SOAP_OK;
}


int srm1__srmStatusOfPutRequest(struct soap* sp, srm1__srmStatusOfPutRequestRequest *req, struct srm1__srmStatusOfPutRequestResponse_ &resp) {
  resp.srmStatusOfPutRequestResponse = soap_new_srm1__srmStatusOfPutRequestResponse(sp,-1);
  resp.srmStatusOfPutRequestResponse->returnStatus=soap_new_srm1__TReturnStatus(sp,-1);
  resp.srmStatusOfPutRequestResponse->returnStatus->statusCode=SRM_USCORENOT_USCORESUPPORTED;
  return SOAP_OK;
}


int srm1__srmStatusOfCopyRequest(struct soap* sp, srm1__srmStatusOfCopyRequestRequest *req, struct srm1__srmStatusOfCopyRequestResponse_ &resp) {
  resp.srmStatusOfCopyRequestResponse = soap_new_srm1__srmStatusOfCopyRequestResponse(sp,-1);
  resp.srmStatusOfCopyRequestResponse->returnStatus=soap_new_srm1__TReturnStatus(sp,-1);
  resp.srmStatusOfCopyRequestResponse->returnStatus->statusCode=SRM_USCORENOT_USCORESUPPORTED;
  return SOAP_OK;
}


int srm1__srmGetRequestSummary(struct soap* sp, srm1__srmGetRequestSummaryRequest *req, struct srm1__srmGetRequestSummaryResponse_ &resp) {
  resp.srmGetRequestSummaryResponse = soap_new_srm1__srmGetRequestSummaryResponse(sp,-1);
  resp.srmGetRequestSummaryResponse->returnStatus=soap_new_srm1__TReturnStatus(sp,-1);
  resp.srmGetRequestSummaryResponse->returnStatus->statusCode=SRM_USCORENOT_USCORESUPPORTED;
  return SOAP_OK;
}


int srm1__srmExtendFileLifeTime(struct soap* sp, srm1__srmExtendFileLifeTimeRequest *req, struct srm1__srmExtendFileLifeTimeResponse_ &resp) {
  resp.srmExtendFileLifeTimeResponse = soap_new_srm1__srmExtendFileLifeTimeResponse(sp,-1);
  resp.srmExtendFileLifeTimeResponse->returnStatus=soap_new_srm1__TReturnStatus(sp,-1);
  resp.srmExtendFileLifeTimeResponse->returnStatus->statusCode=SRM_USCORENOT_USCORESUPPORTED;
  return SOAP_OK;
}


int srm1__srmGetRequestID(struct soap* sp, srm1__srmGetRequestIDRequest *req, struct srm1__srmGetRequestIDResponse_ &resp) {
  resp.srmGetRequestIDResponse = soap_new_srm1__srmGetRequestIDResponse(sp,-1);
  resp.srmGetRequestIDResponse->returnStatus=soap_new_srm1__TReturnStatus(sp,-1);
  resp.srmGetRequestIDResponse->returnStatus->statusCode=SRM_USCORENOT_USCORESUPPORTED;
  return SOAP_OK;
}

// Local

class HTTP_SRM_Handle {
 public:
  string se;
  HTTP_SRM_Handle(const char* se_):se(se_) { };
  ~HTTP_SRM_Handle(void);
};


// Supported methods

int srm1__srmLs(struct soap* sp, srm1__srmLsRequest *req, struct srm1__srmLsResponse_ &resp) {
  HTTP_SRM* it = (HTTP_SRM*)(sp->user);
  // Initialize default error response
  resp.srmLsResponse=soap_new_srm1__srmLsResponse(sp,-1);
  resp.srmLsResponse->returnStatus=soap_new_srm1__TReturnStatus(sp,-1);
  resp.srmLsResponse->returnStatus->statusCode=SRM_USCORENOT_USCORESUPPORTED;
  if((it == NULL) || (it->se == NULL)) {
    resp.srmLsResponse->returnStatus->statusCode=SRM_USCOREFAILURE;
    return SOAP_OK;
  };
  // Find names of requested objects
  if((req == NULL) || (req->paths == NULL) || 
     (req->paths->__ptr == NULL) || (req->paths->__size == 0)) {
    // Nothing requested
    resp.srmLsResponse->returnStatus->statusCode=
                                     SRM_USCOREINVALID_USCOREREQUEST;
    return SOAP_OK;
  };
  if((req->fileStorageType) && (*(req->fileStorageType) != Permanent_)) {
    // Only permanent files are stored in SSE
    resp.srmLsResponse->returnStatus->statusCode=SRM_USCORESUCCESS;
    return SOAP_OK;
  };
  resp.srmLsResponse->details=soap_new_ArrayOfTMetaDataPathDetail(sp,-1);
  resp.srmLsResponse->details->__size=0;
  resp.srmLsResponse->details->__ptr=NULL;
  int n_out = 0;
  int n_alloc = 0;
  for(int n=0;n<req->paths->__size;n++) {
    if(req->paths->__ptr[n] == NULL) continue;
    if(req->paths->__ptr[n]->SURLOrStFN == NULL) continue;
    if(req->paths->__ptr[n]->SURLOrStFN->value == NULL) continue;
    // Create SSE request
    string pattern = req->paths->__ptr[n]->SURLOrStFN->value;
    // SSE has flat structure. Hence either root or file name (so far).
    if(pattern == "/") { pattern="^.*$"; } else { pattern="^"+pattern+"$"; };
    // Pass to SSE
    ns__infoResponse r;
    int res = ns__info((struct soap*)(it->se_sp),(char*)(pattern.c_str()),r);
    // Create SRM response
    if(res != SOAP_OK) {
      // Normally this will never happen
      resp.srmLsResponse->returnStatus->statusCode=SRM_USCOREFAILURE;
      return SOAP_OK;
    };
    if(r.error_code != 0) {
      // TODO: more informative mapping 
      resp.srmLsResponse->returnStatus->statusCode=SRM_USCOREFAILURE;
      return SOAP_OK;
    };
    // Add results to SRM response
    for(int n_info=0;n_info<r.__size_file;n_info++) {
      // new record 
      srm1__TMetaDataPathDetail* rec=soap_new_srm1__TMetaDataPathDetail(sp,-1);
      // manage array's size
      if(n_alloc<=n_out) {
        n_alloc+=10;
        srm1__TMetaDataPathDetail** f_tmp = resp.srmLsResponse->details->__ptr;
        resp.srmLsResponse->details->__ptr=
           (srm1__TMetaDataPathDetail**)soap_malloc(sp,
                          sizeof(srm1__TMetaDataPathDetail*)*n_alloc);
        if(resp.srmLsResponse->details->__ptr == NULL) {
          if(f_tmp != NULL) soap_dealloc(sp,f_tmp);
          resp.srmLsResponse->returnStatus->statusCode=SRM_USCOREFAILURE;
          return SOAP_OK;
        };
        if(f_tmp != NULL) {
          memcpy(resp.srmLsResponse->details->__ptr,
               f_tmp,sizeof(resp.srmLsResponse->details->__ptr)*(n_alloc-10));
          soap_dealloc(sp,f_tmp);
        };
      };
/*
        char *id;
        ULONG64 *size;
        char *checksum;
        char *acl;
        char *created;
        enum ns__filestate *state;
        int __size_url; * sequence of elements <url> *
        char **url;
*/
      // Fill record
      rec->path=r.file[n_info].id;
      rec->status=soap_new_srm1__TReturnStatus(sp,-1); 
      rec->status->statusCode=SRM_USCORESUCCESS;
      if(r.file[n_info].size) {
        rec->size=soap_new_srm1__TSizeInBytes(sp,-1);
        rec->size->value=*(r.file[n_info].size);
      };
      rec->fileStorageType=
       (srm1__TFileStorageType*)soap_malloc(sp,sizeof(srm1__TFileStorageType));
      *(rec->fileStorageType)=Permanent_;
      rec->type=
       (srm1__TFileType*)soap_malloc(sp,sizeof(srm1__TFileType));
      *(rec->type)=File;
      if(r.file[n_info].checksum) {
        char* p = strchr(r.file[n_info].checksum,':');
        if(!p) {
          p=r.file[n_info].checksum;
          rec->checkSumType=soap_new_srm1__TCheckSumType(sp,-1);
          rec->checkSumType->value=soap_strdup(sp,"cksum");
        } else {
          unsigned int l = p-r.file[n_info].checksum;
          rec->checkSumType=soap_new_srm1__TCheckSumType(sp,-1);
          rec->checkSumType->value=(char*)soap_malloc(sp,l+1);
          memcpy(rec->checkSumType->value,r.file[n_info].checksum,l);
          rec->checkSumType->value[l]=0;
          p++; for(;*p;p++) if(!isspace(*p)) break;
        };
        rec->checkSumValue=soap_new_srm1__TCheckSumValue(sp,-1);
        rec->checkSumType->value=soap_strdup(sp,p);
      };
      if(r.file[n_info].created) {
        rec->createdAtTime=soap_new_srm1__TGMTTime(sp,-1);
        rec->lastModificationTime=soap_new_srm1__TGMTTime(sp,-1);
        stringtotime(rec->createdAtTime->value,
                              string(r.file[n_info].created));
        stringtotime(rec->lastModificationTime->value,
                              string(r.file[n_info].created));
      };
      if(r.file[n_info].__size_url) {
        rec->originalSURL=soap_new_srm1__TSURL(sp,-1);
        rec->originalSURL->value=r.file[n_info].url[0];
      };
      resp.srmLsResponse->details->__ptr[n_out]=rec;
/*
      srm1__TOwnerPermission*       ownerPermission  
      ArrayOfTUserPermission*       userPermissions  
      ArrayOfTGroupPermission*      groupPermissions 
      srm1__TOtherPermission*       otherPermission  
      srm1__TUserID*                owner                
      srm1__TLifeTimeInSeconds*     lifetimeAssigned     
      srm1__TLifeTimeInSeconds*     lifetimeLeft         
      srm1__TSURL*                  originalSURL         
      ArrayOfTMetaDataPathDetail*   subPaths             
*/
      n_out++;
    };
  };
  resp.srmLsResponse->details->__size=n_out;
/*
        int *offset;
        int *count;
*/
  return SOAP_OK;
}




// Local service-specific

bool srm_service_configurator(istream& f,const char* uri,HTTP_Service_Properties &prop) {
  string se_path("");
  for(;!f.eof();) {
    char buf[1024];
    istream_readline(f,buf,sizeof(buf));
    char* p = buf;
    for(;*p;p++) if(!isspace(*p)) break;
    char* command = p;
    if(*command == '#') continue;
    for(;*p;p++) if(isspace(*p)) break;
    int command_len = p-command;
    if(!command_len) continue;
    if((command_len==2) && (strncmp(command,"se",command_len)==0)) {
      input_escaped_string(p,se_path);
    } else {
      odlog(0)<<"SRM(configure): skipping unknown command: "<<command<<endl;
    };
  };
  prop.arg=NULL;
  odlog(0)<<"SRM(configure): links to SSE at : "<<se_path<<endl;
  if(se_path.length()) {
    HTTP_SRM_Handle* handle = new HTTP_SRM_Handle(se_path.c_str());
    prop.arg=handle;
  };
  if(!prop.arg) {
    odlog(0)<<"SRM(configure): failed"<<endl;
  };
  prop.subtree=false;
  return true;
}

HTTP_Service* srm_service_creator(HTTPS_Connector& c,const char* uri,void* arg) {
  HTTP_SRM_Handle* handle = (HTTP_SRM_Handle*)arg;
  if(!handle) return NULL;
  HTTP_SRM* h = new HTTP_SRM(&c,handle->se.c_str());
  return h;
}  

extern HTTP_Services services;

HTTP_SRM::HTTP_SRM(HTTPS_Connector *c_,const char* se_):HTTP_ServiceAdv(c_),se(NULL),se_path(se_),c(c_) {
  // if(!c) return;
  soap_init();
  sp.namespaces=srm_soap_namespaces;
  sp.user=this;
}

HTTP_SRM::~HTTP_SRM(void) {
  soap_deinit();
}

HTTP_Error HTTP_SRM::get(const char* uri,int &keep_alive) {
  return HTTP_NOT_IMPLEMENTED;
}

HTTP_Error HTTP_SRM::put(const char* uri,int &keep_alive) {
  return HTTP_NOT_IMPLEMENTED;
}

HTTP_Error HTTP_SRM::post(const char* uri,int &keep_alive) {
  odlog(3)<<"SRM:post: uri: "<<uri<<endl;
  // obtain SSE service to use
  if(se == NULL) {
    if(c) se=services.get(*c,se_path.c_str());
    if(se == NULL) {
      odilog(0,c->pid)<<"Corresponding SSE service not found: "<<se_path<<endl;
      return HTTP_NOT_IMPLEMENTED;
    };
    // TODO: fix "dangerous" cast
    se_sp=&(((HTTP_SRM*)se)->sp);
  };
  return soap_post(uri,keep_alive);
}

