#include "../../std.h"

#include "../../misc/globus_error_utils.h"
#include "../../misc/log_time.h"
 
#include <stdsoap2.h>

#include "client.h"


int HTTP_ClientSOAP::local_fsend(struct soap *sp, const char* buf, size_t l) {
  if(sp->socket != 0) return SOAP_SSL_ERROR;
  HTTP_ClientSOAP *it = (HTTP_ClientSOAP*)(sp->user);
  it->cond_write.reset();
  // do not register read - so cond_write will be set by writer only
  odlog(3)<<"*** Client sends: ";
  for(size_t n=0;n<l;n++) odlog_(3)<<buf[n];
  odlog_(3)<<endl;
  globus_result_t res = globus_io_register_write(&(it->s),(globus_byte_t*)buf,
                      l,&write_callback,it);
  if(res != GLOBUS_SUCCESS) {
    globus_io_cancel(&(it->s),GLOBUS_FALSE); // cancel waiting read
    olog<<res<<endl;
    return SOAP_SSL_ERROR;
  };
  globus_thread_blocking_will_block(); // ????
  int r;
  if(!(it->cond_write.wait(r,it->timeout))) {
    olog<<"Timeout while sending SOAP request"<<endl;
    globus_io_cancel(&(it->s),GLOBUS_FALSE);
    return SOAP_SSL_ERROR;
  };
  if(r != 0) {
    olog<<"Error sending data to server"<<endl;
    return SOAP_SSL_ERROR;
  };
  return SOAP_OK;
}

size_t HTTP_ClientSOAP::local_frecv(struct soap* sp, char* buf, size_t l) {
  if(sp->socket != 0) return 0;
  HTTP_ClientSOAP *it = (HTTP_ClientSOAP*)(sp->user);
  globus_result_t res = globus_io_register_read(&(it->s),(globus_byte_t*)buf,
                                       l,1,&read_callback,it);
  if(res != GLOBUS_SUCCESS) {
    olog<<res<<endl;
    return 0;
  };
  globus_thread_blocking_will_block(); // ????
  int r;
  if(!(it->cond_read.wait(r,it->timeout))) {
    globus_io_cancel(&(it->s),GLOBUS_FALSE); return 0;
  };
  if(r != 0) return 0;
  return it->answer_size;
}

int HTTP_ClientSOAP::local_fopen(struct soap* sp, const char* endpoint, const char* host, int port) {
  HTTP_ClientSOAP *it = (HTTP_ClientSOAP*)(sp->user);
  sp->socket=-1;
  if(sp->socket == 0) {
    olog<<"Multiple soap_fopen"<<endl;
  };
  if(it->connect() != 0) return -1;
  sp->socket=0;
  return 0;
};

int HTTP_ClientSOAP::local_fclose(struct soap* sp) {
  return 0;
};


HTTP_ClientSOAP::HTTP_ClientSOAP(const char* base,struct soap *sp):HTTP_Client(base),soap(sp) {
  /* intialize soap */
  soap_init(soap);
  soap->fsend=&local_fsend;
  soap->frecv=&local_frecv;
  soap->fopen=&local_fopen;
  soap->fclose=&local_fclose;
  soap->http_version="1.1";
  soap->keep_alive=1;
  soap_set_imode(soap,SOAP_IO_KEEPALIVE);
  soap_set_omode(soap,SOAP_IO_KEEPALIVE);
  soap->user=this;
  soap_url=base;
  int n = soap_url.find(':');
  if(n != string::npos) soap_url.replace(0,n,"http"); // fake url for gSOAP
}  

HTTP_ClientSOAP::~HTTP_ClientSOAP(void) {
  soap_destroy(soap);
  soap_end(soap);
  soap_done(soap);
}

void HTTP_ClientSOAP::reset(void) {
  soap_destroy(soap);
  soap_end(soap);
}

