#ifndef FLO_COMMON_H
#define FLO_COMMON_H

#include "condition.h"

class pSafeObject;

class SafeObject {
 friend class pSafeObject;
 private:
  int p_counter_;
  LockSimple* p_lock_;
  void acquire(void) {
    p_lock_->block();
    p_counter_++;
    p_lock_->unblock();
  };
  void release(void) {
    LockSimple* lock = p_lock_;
    p_lock_->block();
    if((--p_counter_) == 0) delete this;
    lock->unblock();
    delete this;
  };
 public:
  SafeObject(void):p_counter_(0) { p_lock_ = new LockSimple(); };
  virtual ~SafeObject(void) { };
};

class pSafeObject {
 protected:
  SafeObject* object_;
 public:
  pSafeObject(void) { object_=NULL; };
  pSafeObject(SafeObject& o) { o.acquire(); object_=&o; };
  pSafeObject(SafeObject* o) { if(o) o->acquire(); object_=o; };
  ~pSafeObject(void) { if(object_) object_->release(); };
  SafeObject& operator*(void) const { return *object_; };
  SafeObject* operator->(void) const { return object_; };
  void operator=(SafeObject& o) {
    o.acquire();
    if(object_) object_->release();
    object_=&o;
  };
  void operator=(SafeObject* o) {
    if(o) o->acquire();
    if(object_) object_->release();
    object_=o;
  };
  void operator=(const pSafeObject& p) {
    if(p.object_) p.object_->acquire();
    if(object_) object_->release();
    object_=p.object_;
  };
  operator bool(void) const { return (object_!=NULL); };
  bool operator!(void) const { return (object_==NULL); };
  bool operator==(const pSafeObject& p) const { return (object_ == p.object_); };
  bool operator!=(const pSafeObject& p) const { return (object_ != p.object_); };
  bool operator==(const SafeObject& o) const { return (object_ == &o); };
  bool operator!=(const SafeObject& o) const { return (object_ != &o); };
};

#endif // FLO_COMMON_H
