#ifndef SERVICE_H
#define SERVICE_H

#include "common.h"
#include "pointers.h"

#include "message.h"

// Interface to remote registered service
class FlOServiceRemote: public SafeObject {
 private:
  // URL to contact service
  std::string url_;
  // DN of certificate used to identfy service
  std::string id_;
  // Queue of messages
  list<FlOMessage> queue_;
  LockSimple queue_lock_;
  typedef enum {
    unknown
  } state_t;
  bool acceptsCreatedOrdinary;
  bool acceptsCreatedReference;
  bool acceptsUpdatedOrdinary;
  bool acceptsUpdatedReference;
  bool acceptsDeletedOrdinary;
  bool acceptsDeletedReference;
 public:
  FlOServiceRemote(const std::string& url,const std::string& id);
  ~FlOServiceRemote(void);
  const std::string& url(void) { return url_; };
  const std::string& id(void) { return id_; };
  // Put messages into a queue (another implementation can send it immediately)
  void notify(const FlOObject& o,FlOObject::action_t t);
  void maintain(void);
};

// Kind of main class. Represents local service itself.
class FlOService {
 private:
  bool valid;
  // List or registered services
  list<pFlOServiceRemote> services;
  // List of handled objects
  list<pFlOObject> objects;
 public:
  FlOService(void);
  operator bool(void) { return valid; };
  bool operator!(void) { return !valid; };
  void add(pFlOObject o);
  // Inform registered services about changes
  void notify(pFlOObject o,FlOObject::action_t t);
  // Accept notification from remote service
  void notification(FlOMessage& msg,pFlOServiceRemote& srv);
  pFlOServiceRemote service(const std::string& id);
};

#endif // SERVICE_H

