#ifndef __HTTPSD_H__
#define __HTTPSD_H__

#include <globus_io.h>
#include <list>
#include <string>

#include "../auth/auth.h"

class HTTP_Service;
class HTTP_Services;
class AuthEvaluator;

class HTTPS_Connector {
 private:
  // active connections
  static list<HTTPS_Connector*> cons;
  static pthread_mutex_t cons_lock;
  bool initialized;
  // handle for data transfer
  globus_io_handle_t h;
  string base_url;
  // configured services
  HTTP_Services& services;
  // confugured authoziration rules
  list<AuthEvaluator*>& auths;
//  string subject;
  // authencity of connected object
  AuthUser user;
  static void accept_cb(void *callback_arg,globus_io_handle_t *handle,globus_result_t result);
  char r_buf[65536]; // better if it is enough for whole header
  unsigned int r_start;
  unsigned int r_end;
  static void* loop_thread(void*);
 public:
  unsigned int pid;
  HTTPS_Connector(globus_io_handle_t *s,const char* url,HTTP_Services& serv,list<AuthEvaluator*>& auths);
  ~HTTPS_Connector(void);
  operator bool(void) { return initialized; };
  size_t read(char* buf, size_t l);
  int write(const char* buf, size_t l);
  size_t readline(char* buf, size_t l);
//  bool add_service(const char *url,HTTP_Service *service);
//  bool remove_service(const char *url);
  void loop(void);
  void loop_in_thread(void);
  static void identity(globus_io_handle_t* handle,const char* subject);
  const char* identity_subject(void) const { return user.DN(); };
  const char* identity_proxy(void) const { return user.proxy(); };
  AuthUser& identity(void) { return user; };
  list<AuthEvaluator*>& authorizations(void) { return auths; };
  const char* url(void) { return base_url.c_str(); };
};

typedef enum {
  HTTP_OK = 200,
  HTTP_NOT_IMPLEMENTED = 501,
  HTTP_NOT_ALLOWED = 403,
  HTTP_NOT_FOUND = 404,
  HTTP_ERROR = 500,
  HTTP_FAILURE = -1
} HTTP_Error;

int skip_header(HTTPS_Connector &c,int &keep_alive);
int skip_request(HTTPS_Connector &c,int &keep_alive);
int send_response_header(HTTPS_Connector &c,int keep_alive,int code,char* type,int size);
int send_file(const char* fname,HTTPS_Connector &c);
int stat_file(const char* fname,unsigned long long int &size);
int send_error_response(HTTPS_Connector &c,int keep_alive,int code,char* type,char* content);

// tempate for any service accessible through Connector
class HTTP_Service {
 public:
  HTTP_Service(void);
  virtual ~HTTP_Service(void);
  virtual HTTP_Error get(const char* uri,int &keep_alive);
  virtual HTTP_Error put(const char* uri,int &keep_alive);
  virtual HTTP_Error post(const char* uri,int &keep_alive);
};

class HTTP_Service_Properties {
 public:
  // whole is subtree is served by this service
  bool subtree;
  // configuration-dependent information passed during configuration
  void* arg;
};

// each service class must have such corresponding functions
typedef HTTP_Service* (*service_creator)(HTTPS_Connector& c,const char* uri,void* arg);
typedef bool (*service_configurator)(istream& f,const char* uri,HTTP_Service_Properties &prop);

typedef struct {
  const char* name;
  service_creator      creat;
  service_configurator config;
} service_item_t;

// description of service created during configuration one per service
class HTTP_Service_description {
 friend class HTTP_Services;
 private:
  // URL served by this service
  string uri;
  service_creator creator;
  HTTP_Service_Properties properties;
 public:
  HTTP_Service_description(const char* u,service_creator c,const HTTP_Service_Properties &prop);
  // return true if 'u' contains url served by this service
  bool match(const char* u);
  HTTP_Service* create(HTTPS_Connector& c);
};

class HTTP_Service_item {
 public:
  const HTTPS_Connector* con;
  // URL for which this service instance was created
  string uri;
  bool subtree;
  HTTP_Service* serv;
  HTTP_Service_item(const HTTPS_Connector* c,const char* u,bool st,HTTP_Service* s):con(c),uri(u),subtree(st),serv(s) { };
  ~HTTP_Service_item(void) { };
  bool match(const char *u);
};

// Holds list of services and manages their creation and destruction
class HTTP_Services {
 private:
  list<HTTP_Service_description> desc;
  list<HTTP_Service_item> serv;
  pthread_mutex_t lock;
 public:
  HTTP_Services(void);
  ~HTTP_Services(void);
  bool add(const char* u,service_creator c,const HTTP_Service_Properties &prop);
  bool remove(const char* u,service_creator c);
  HTTP_Service* get(HTTPS_Connector& c,const char* u);
  void close(const HTTPS_Connector& c);
};

#endif /* __HTTPSD_H__ */
