// TODO: Run multiple clients in parallel

#include "../../std.h"
#include <globus_common.h>
#include <globus_io.h>
#include <fstream>

#include "../../files/info_types.h"
#include "../../misc/log_time.h"
#include "../../misc/inttostring.h"
#include "../../misc/url_options.h"
#include "../../misc/globus_error_utils.h"

#include "../client/client.h"
#include "../../misc/url_options.h"

#include "logger_soap.nsmap"

#define copy_str(to,from,member) { \
  to.member=NULL; if(from.member) to.member=strdup(from.member); \
}

void copy_jobinfo(nl__jobinfo& to,const nl__jobinfo& from) {
  to.start=from.start;
  to.end=from.end;
  copy_str(to,from,cluster);
  copy_str(to,from,user);
  copy_str(to,from,id);
  copy_str(to,from,name);
  copy_str(to,from,failure);
  copy_str(to,from,lrms);
  copy_str(to,from,queue);
  copy_str(to,from,rsl);
  copy_str(to,from,ui);
  to.usedcpu=from.usedcpu;
  to.usedmem=from.usedmem;
};

//struct Namespace namespaces[] =
//{
//  {NULL, NULL}
//};

class LoggerClient {
 private:
  URL url;
  HTTP_ClientSOAP* client;
  struct soap soap;
 public:
  LoggerClient(void);
  ~LoggerClient(void);
  bool Initialized(void);
  bool SameContact(const char* url);
  bool Report(const char* url,nl__jobinfo& info);
  bool Query(const char* url,const char* q,unsigned long long offset,list<nl__jobinfo>& info);
  bool Query(const char* url,const char* q,unsigned long long offset,array_jobinfo* info);
};


LoggerClient::LoggerClient(void):url(NULL),client(NULL) {
}

LoggerClient::~LoggerClient(void) {
  if(client) delete client;
}

bool LoggerClient::Initialized(void) {
  return (client != NULL);
}

bool LoggerClient::SameContact(const char* url_) {
  if(!url) return false;
  URL u(url_);
  return((u.Proto() == url.Proto()) && 
         (u.Host() == url.Host()) && 
         (u.Port() == url.Port()));
}

bool LoggerClient::Report(const char* url_,nl__jobinfo& info) {
  URL u(url_);
  if(url) {
    if((u.Proto() == url.Proto()) &&
       (u.Host() == url.Host()) &&
       (u.Port() == url.Port())) {
      url=u;
    } else { // reinitialize for different server
      url=u; if(client) delete client; client=NULL;
    };
  } else { // initialize for first connection
    url=u; if(client) delete client; client=NULL;
  };
  if(client == NULL) {
    string soap_url = url.Proto()+"://"+url.Host()+":"+inttostring(url.Port());
    client = new HTTP_ClientSOAP(soap_url.c_str(),&soap);
    soap.namespaces=logger_soap_namespaces;
  } else {
    client->reset();
  };
  // connect to server if not yet connected
  if(client->connect() != 0) {
    odlog(1)<<"Failed to connect to "<<url<<endl;
    return false;
  };
  // Send report
  string fake_soap_url = string("http://")+url.Host()+":"+inttostring(url.Port())+"/"+url.Path();
  int soap_err = SOAP_OK;
  int r;
  if((soap_err=soap_call_nl__add(&soap,fake_soap_url.c_str(),"add",&info,r)) != SOAP_OK) {
    odlog(1)<<"Failed to pass information to database"<<endl;
    r=1;
    if(LogTime::level > 0) soap_print_fault(&soap, stderr);
    client->disconnect();
  } else if(r != 0) {
    odlog(1)<<"Record refused. Error code: "<<r<<endl;
  };
  // Do not disconnect
  return (r==0);
}

bool LoggerClient::Query(const char* url_,const char* q,unsigned long long offset,list<nl__jobinfo>& info) {
  array_jobinfo r;
  bool res = Query(url_,q,offset,&r);
  if(!res) return false;
  for(int i = 0; i<r.__size_job; i++) {
    nl__jobinfo job;
    copy_jobinfo(job,r.job[i]);
    info.push_back(job);
  };
  return true;
}

bool LoggerClient::Query(const char* url_,const char* q,unsigned long long offset,array_jobinfo* info) {
  info->__size_job=0;
  info->job=NULL;
  if(url_ == NULL) return false;
  URL u(url_);
  if(url) {
    if((u.Proto() == url.Proto()) &&
       (u.Host() == url.Host()) &&
       (u.Port() == url.Port())) {
      url=u;
    } else { // reinitialize for different server
      url=u; if(client) delete client; client=NULL;
    };
  } else { // initialize for first connection
    url=u; if(client) delete client; client=NULL;
  };
  if(client == NULL) {
    string soap_url = url.Proto()+"://"+url.Host()+":"+inttostring(url.Port());
    client = new HTTP_ClientSOAP(soap_url.c_str(),&soap);
    soap.namespaces=logger_soap_namespaces;
  } else {
    client->reset();
  };
  // connect to server if not yet connected
  if(client->connect() != 0) {
    odlog(1)<<"Failed to connect to "<<url<<endl;
    return false;
  };
  // Query server
  string fake_soap_url = string("http://")+url.Host()+":"+inttostring(url.Port())+"/"+url.Path();
  int soap_err = SOAP_OK;
  array_jobinfo r;
  if((soap_err=soap_call_nl__get(&soap,fake_soap_url.c_str(),"get",(char*)q,offset,r)) != SOAP_OK) {
    odlog(1)<<"Failed to get information from database"<<endl;
    if(LogTime::level > 0) soap_print_fault(&soap, stderr);
    client->disconnect();
    return false;
  };
  (*info)=r;
  // Do not disconnect
  return true;
}

#define MAX_CLIENTS 10
HTTP_ClientSOAP* clients[MAX_CLIENTS];

typedef struct {
  bool start,end,cluster,user,id,name,failure,lrms,queue,rsl,ui,usedcpu,usedmem;
} elements_t;

#define set_element(v,e,name) if(strcmp(name,v) == 0) { e=true; }

int logger(const char* url,char const * const * dirs,int num,time_t ex_period = 0);
int query(const char* url,char const * q,unsigned long int offset,unsigned int size,bool xml_out,const elements_t& e);
 
int main(int argc,char* argv[]) {
  char* url = NULL;
  elements_t e = { true, true, true, true, true, true, true, true, true, true, true, true, true };
  time_t ex_period = 0;

  opterr=0;
  bool q = false;
  bool xml_out = false;
  unsigned long int offset = 0;
  unsigned long int size = 100;
  int n;
  while((n=getopt(argc,argv,":hxe:qu:d:o:s:E:")) != -1) {
    switch(n) {
      case ':': { olog<<"Missing argument\n"; return 1; };
      case '?': { olog<<"Unrecognized option\n"; return 1; };
      case 'h': {
        cout<<"logger [-h] [-d level] [-u url] {control_dir ...|-q [-x] [-o offset] [-s size] [-e element,element,...] [-E expiration_period_days] [query]}"<<endl;
         return 0;
      };
      case 'u': {
        url=optarg;
      }; break;
      case 'q': {
        q=true;
      }; break;
      case 'd': {
        char* p;
        int i = strtol(optarg,&p,10);
        if(((*p) != 0) || (i<0)) {
          olog<<"Improper debug level '"<<optarg<<"'"<<endl;
          exit(1);
        };
        LogTime::level=i;
      }; break;
      case 'e': {
        e.start=false; e.end=false; e.cluster=false; e.user=false;
        e.id=false; e.name=false; e.failure=false; e.lrms=false;
        e.queue=false; e.rsl=false; e.ui=false; e.usedcpu=false;
        e.usedmem=false;
        for(const char* p=optarg;*p;) {
          const char* pp = strchr(p,',');
          string name;
          if(pp) { name.assign(p,pp-p); } else { name.assign(p); };
          set_element(name.c_str(),e.start,"start")
          else set_element(name.c_str(),e.end,"end")
          else set_element(name.c_str(),e.cluster,"cluster")
          else set_element(name.c_str(),e.user,"user")
          else set_element(name.c_str(),e.id,"id")
          else set_element(name.c_str(),e.name,"name")
          else set_element(name.c_str(),e.failure,"failure")
          else set_element(name.c_str(),e.lrms,"lrms")
          else set_element(name.c_str(),e.queue,"queue")
          else set_element(name.c_str(),e.ui,"ui")
          else set_element(name.c_str(),e.usedcpu,"usedcpu")
          else set_element(name.c_str(),e.usedmem,"usedmem");
          if(!pp) break;
          p=pp+1;
        };
      }; break;
      case 'o': {
        char* p;
        int i = strtol(optarg,&p,10);
        if(((*p) != 0) || (i<0)) {
          olog<<"Improper offset '"<<optarg<<"'"<<endl;
          exit(1);
        };
        offset=i;
      }; break;
      case 's': {
        char* p;
        int i = strtol(optarg,&p,10);
        if(((*p) != 0) || (i<=0)) {
          olog<<"Improper size '"<<optarg<<"'"<<endl;
          exit(1);
        };
        size=i;
      }; break;
      case 'E': {
        char* p;
        int i = strtol(optarg,&p,10);
        if(((*p) != 0) || (i<=0)) {
          olog<<"Improper expiration period '"<<optarg<<"'"<<endl;
          exit(1);
        };
        ex_period=i; ex_period*=(60*60*24);
      }; break;
      case 'x': {
        xml_out=true;
      }; break;
      default: { olog<<"Options processing error\n"; return 1; };
    };
  };
  if(!q) return logger(url,argv+optind,argc-optind,ex_period);
  return query(url,argv[optind],offset,size,xml_out,e);
}

int logger(const char* url,char const * const * dirs,int num,time_t ex_period) {
  int ret = 0;
  if(globus_module_activate(GLOBUS_IO_MODULE) != GLOBUS_SUCCESS) exit(1);

  {
  LoggerClient client;
  for(int n = 0;dirs[n] && n<num;n++) {
    string logdir(dirs[n]); logdir+="/logs/";
    struct dirent file_;
    struct dirent *file;
    odlog(1)<<"Processing directory: "<<logdir<<endl;
    DIR *dir=opendir(logdir.c_str());
    if(dir == NULL) continue;
    for(;;) {
      readdir_r(dir,&file_,&file);
      if(file == NULL) break;
      string logfile = logdir+file->d_name;
      struct stat st;
      if(stat(logfile.c_str(),&st) != 0) continue;
      if(!S_ISREG(st.st_mode)) continue;
      odlog(1)<<"Processing file: "<<logfile<<endl;
      /* read file */
      char buf[4096];
      ifstream f(logfile.c_str());
      if(!f.is_open() ) continue; /* can't open file */
      nl__jobinfo info;  // TODO: constructor/destructor
      info.start=0; info.end=0; info.cluster=NULL; info.user=NULL;
      info.id=NULL; info.name=NULL; info.failure=NULL; info.lrms=NULL;
      info.queue=NULL; info.rsl=NULL; info.ui=NULL;
      info.usedcpu=0; info.usedmem=0;
      char* job_url=NULL;
      const char* logger_url=url;
      for(;!f.eof();) {
        istream_readline(f,buf,sizeof(buf));
        char* value = strchr(buf,'=');
        if(!value) continue;
        char* name=buf; (*value)=0; value++;
        #define COMPARE_AND_SET(NM,ATTR) \
          if(!strcmp(name,NM)) { \
            if(info.ATTR) free(info.ATTR); \
            info.ATTR=strdup(value); continue; \
          }
        odlog(2)<<"name: "<<name<<" - value: "<<value<<endl;
        if(!strcmp(name,"loggerurl")) {
          if(job_url) free(job_url); job_url=strdup(value); continue; 
        };
        if(!strcmp(name,"starttime")) {
          mds_time t; t=value; if(t != (time_t)(-1)) info.start=t;
          continue; 
        };
        if(!strcmp(name,"endtime")) {
          mds_time t; t=value; if(t != (time_t)(-1)) info.end=t;
          continue; 
        };
	if(!strcmp(name,"walltime")) {
	  info.usedcpu = atoi(value);
	  continue;
	}
        COMPARE_AND_SET("gridid",id);
        COMPARE_AND_SET("lrms",lrms);
        COMPARE_AND_SET("queue",queue);
        COMPARE_AND_SET("subject",user);
        COMPARE_AND_SET("jobname",name);
        COMPARE_AND_SET("failure",failure);
        COMPARE_AND_SET("rsl",rsl);
        COMPARE_AND_SET("clientname",ui);
      };
      f.close();
      if(job_url) logger_url=job_url;
      bool result = true;
      if(logger_url == NULL) {
        odlog(2)<<"No service URL provided"<<endl;
        result = false;
      } else {
        odlog(2)<<"Reporting to: "<<logger_url<<endl;
        result = client.Report(logger_url,info);
      };
      if(result) {
        odlog(0)<<"Passed information about job "<<info.id<<endl;
        unlink(logfile.c_str());
      } else {
        odlog(0)<<"Failed to pass information about job "<<info.id<<endl;
        ret=-1;
        // Check creation time and remove it if really too old
        if((ex_period) && 
           (((unsigned int)(time(NULL)-st.st_mtime)) > ex_period)) {
          odlog(0)<<"Removing outdated information about job "<<info.id<<endl;
          unlink(logfile.c_str());
        };
      };
      if(info.cluster) free(info.cluster);
      if(info.user) free(info.user);
      if(info.id) free(info.id);
      if(info.name) free(info.name);
      if(info.failure) free(info.failure);
      if(info.lrms) free(info.lrms);
      if(info.queue) free(info.queue);
      if(info.rsl) free(info.rsl);
      if(info.ui) free(info.ui);
      if(job_url) free(job_url);
    };
    closedir(dir);
  };
  };
  globus_module_deactivate(GLOBUS_IO_MODULE);
  return ret;
}

int query(const char* url,char const * q,unsigned long int offset,unsigned int size,bool xml_out,const elements_t& e) {
  if(url == NULL) {
    odlog(0)<<"URL is needed"<<endl;
    return 1;
  };
  if(globus_module_activate(GLOBUS_IO_MODULE) != GLOBUS_SUCCESS) exit(1);
  list<nl__jobinfo> info;
  bool res = true;
  unsigned long int o = offset;
  unsigned int s = 0;
  nl__jobinfo* a = (nl__jobinfo*)malloc(sizeof(nl__jobinfo)*size);
  if(a == NULL) {
    odlog(0)<<"Failed to allocate memory"<<endl;
    globus_module_deactivate(GLOBUS_IO_MODULE);
    return 1;
  };
  {
    LoggerClient client;
    array_jobinfo info;
    for(;s<size;) {
      res = client.Query(url,q,o,&info);
      if(!res) break;
      if(info.__size_job == 0) break;
      if(info.__size_job > (size-s)) info.__size_job=size-s;
      memcpy(a+s,info.job,sizeof(nl__jobinfo)*info.__size_job);
      s+=info.__size_job; o+=info.__size_job;
    };
    if(!res && !s) {
      odlog(0)<<"Failed to retrieve information"<<endl;
    } else if(s) {
      if(!res) odlog(0)<<"Warning: not full information was retrieved"<<endl;
      res=true;
      nl__jobinfo* p = a;
      int i = 0;
      for(;i<s;i++,p++) {
        if(!xml_out) {
          if(!e.start) p->start=0;
          if(!e.end) p->end=0;
          if(!e.usedcpu) p->usedcpu=0;
          if(!e.usedmem) p->usedmem=0;
        };
        if(!e.cluster) p->cluster=NULL;
        if(!e.user) p->user=NULL;
        if(!e.id) p->id=NULL;
        if(!e.name) p->name=NULL;
        if(!e.failure) p->failure=NULL;
        if(!e.lrms) p->lrms=NULL;
        if(!e.queue) p->queue=NULL;
        if(!e.rsl) p->rsl=NULL;
        if(!e.ui) p->ui=NULL;
      };
      if(!xml_out) {
        nl__jobinfo* p = a;
        int i = 0;
        for(;i<s;i++,p++) {
          if(p->id) cout<<"id: "<<p->id<<endl;
          if(p->cluster) cout<<"cluster: "<<p->cluster<<endl;
          if(p->user) cout<<"user: "<<p->user<<endl;
          if(p->ui) cout<<"ui: "<<p->ui<<endl;
          if(p->name) cout<<"name: "<<p->name<<endl;
          if(p->start) cout<<"start: "<<p->start<<endl;
          if(p->end) cout<<"end: "<<p->end<<endl;
          if(p->failure) cout<<"failure: "<<p->failure<<endl;
          if(p->lrms) cout<<"lrms: "<<p->lrms<<endl;
          if(p->queue) cout<<"queue: "<<p->queue<<endl;
          if(p->usedcpu) cout<<"usedcpu: "<<p->usedcpu<<endl;
          if(p->usedmem) cout<<"usedmem: "<<p->usedmem<<endl;
          if(p->rsl) cout<<"rsl: "<<p->rsl<<endl;
          cout<<endl;
        };
      } else { // serialize using gSOAP
        array_jobinfo ar;
        ar.__size_job=s;
        ar.job=a;
        struct soap sp;
        soap_init(&sp);
        soap_begin(&sp);
        soap_set_omode(&sp,SOAP_XML_GRAPH | SOAP_XML_CANONICAL);
        ar.soap_serialize(&sp);
        sp.os=&cout;
        soap_begin_send(&sp);
        ar.soap_put(&sp,"id","array-jobinfo");
        soap_end_send(&sp);
        soap_end(&sp);
        soap_done(&sp);
      };
    } else {
      odlog(0)<<"Warning: no data was retrieved"<<endl;
    };
  };
  globus_module_deactivate(GLOBUS_IO_MODULE);
  if(!res) return 1;
  return 0;
}
