#ifndef __SE_FILE_H__
#define __SE_FILE_H__

#include "../../std.h"

#include <string>
#include <list>
#include <iostream>

#include "disk_space.h"
#include "se_ns.h"
#include "../../misc/condition.h"
//#include "../../auth/auth.h"
#include "../../auth/gacl_auth.h"
#include "../../misc/safelist.h"

/*
extern "C" {
#ifdef NG_GACL
#include <gacl.h>
typedef struct _GACLnamevalue GACLnamevalue;
#else
#ifdef GRIDSITE_GACL
#define GACLnamevalue GRSTgaclNamevalue
#include <gridsite.h>
#include <gridsite-gacl.h>
#endif
#endif
}
*/

/*
class SEUserDescription {
 public:
  string subject;
  list<string> group; // any information used to group users presented in one line
};
*/

// Access control
#define FILE_ACC_NONE        GACL_PERM_NONE
#define FILE_ACC_LIST        GACL_PERM_LIST
#define FILE_ACC_READ        GACL_PERM_READ
#define FILE_ACC_WRITE       GACL_PERM_WRITE
#define FILE_ACC_CREATE      GACL_PERM_WRITE
#define FILE_ACC_DELETE      GACL_PERM_WRITE
#define FILE_ACC_READ_META   GACL_PERM_ADMIN
#define FILE_ACC_WRITE_META  GACL_PERM_ADMIN
#define FILE_ACC_ALL         0xFFFF 

#define FILE_STATE_ACCEPTED       0    // File accepted at SE (initial state)
#define FILE_STATE_COLLECTING     1    // File is being uploaded by client
#define FILE_STATE_REQUESTED      2    // File should be downloaded by service
#define FILE_STATE_DOWNLOADING    3    // File is being downloaded
#define FILE_STATE_COMPLETE       4    // File is physically complete
#define FILE_STATE_VALID          5    // File's content is valid
#define FILE_STATE_DELETING       6    // File is being deleted

#define FILE_STATE_MAX            7
#define FILE_STATE_UNCHANGED      FILE_STATE_MAX

#define REG_STATE_LOCAL           0
#define REG_STATE_REGISTERING     1  // File claimed - registration in progress
#define REG_STATE_ANNOUNCED       2
#define REG_STATE_UNREGISTERING   3  // File claimed - registration in progress

#define REG_STATE_MAX             4
#define REG_STATE_UNCHANGED       REG_STATE_MAX

extern const char* file_state_str[];
extern const char* reg_state_str[];

#define FILE_STATE_ACCEPTED_STR      file_state_str[FILE_STATE_ACCEPTED]
#define FILE_STATE_COLLECTING_STR    file_state_str[FILE_STATE_COLLECTING
#define FILE_STATE_COMPLETE_STR      file_state_str[FILE_STATE_COMPLETE]
#define FILE_STATE_VALID_STR         file_state_str[FILE_STATE_COMPLETE]
#define FILE_STATE_DELETING_STR      file_state_str[FILE_STATE_DELETING]

#define REG_STATE_LOCAL_STR          reg_state_str[REG_STATE_LOCAL]
#define REG_STATE_REGISTERING_STR    reg_state_str[REG_STATE_REGISTERING]
#define REG_STATE_ANNOUNCED_STR      reg_state_str[REG_STATE_ANNOUNCED]
#define REG_STATE_UNREGISTERING_STR  reg_state_str[REG_STATE_UNREGISTERING]



typedef struct {
  uint64_t start;
  uint64_t end;
} SEFileRange;
#define MAX_SEFILE_RANGES 100
#define MAX_RANGE_VALUE ((uint64_t)(-1))
#define MIN_RANGE_VALUE (0)

class SEAuthorization {

};

// File metadata
class SEAttributes {
 private:
  bool valid;
  uint64_t size_i;
  bool size_b;
  string id_i;
  string creator_i;
  string checksum_i;
  bool checksum_b;
  struct tm created_i;
  bool created_b;
  list<string> sources_;
 public:
  operator bool(void) { return valid; };
  bool operator!(void) { return !valid; };
  bool enough(void);
  bool complete(void);
  SEAttributes(void):valid(false),size_b(false),checksum_b(false),created_b(false) { };
  SEAttributes(const char* id,uint64_t size,const AuthUser user):id_i(id),size_b(true),size_i(size),valid(true) {
    checksum_b=false;
    created_b=false;
    creator_i=user.DN();
  };
  SEAttributes(const char* id,const AuthUser user):id_i(id),size_b(false),valid(false) {
    checksum_b=false;
    created_b=false;
    creator_i=user.DN();
  };
  SEAttributes(const SEAttributes& attr):valid(attr.size_b) {
    id_i=attr.id_i;
    size_i=attr.size_i; size_b=attr.size_b;
    if((checksum_b=attr.checksum_b)) checksum_i=attr.checksum_i;
    if((created_b=attr.created_b)) memcpy(&created_i,&(attr.created_i),sizeof(created_i));
    creator_i=attr.creator_i;
    sources_=attr.sources_;
  };
  SEAttributes(const char* fname);
  void operator=(const SEAttributes& attr) {
    valid=attr.valid;
    id_i=attr.id_i;
    size_i=attr.size_i; size_b=attr.size_b;
    checksum_b=attr.checksum_b; checksum_i=attr.checksum_i;
    created_b=attr.created_b; memcpy(&created_i,&(attr.created_i),sizeof(created_i));
    creator_i=attr.creator_i;
    sources_=attr.sources_;
  };
  int read(const char* fname);
  int write(const char* fname);
  uint64_t size(void) const {
    if(!size_b) return MAX_RANGE_VALUE;
    return size_i;
  };
  void size(uint64_t s) { size_i=s; size_b=true; valid=true; };
  bool size_available(void) { return size_b; };
  const char* id(void) const { return id_i.c_str(); };
  void id(const char* i) { id_i=i; };
  const string& creator(void) const { return creator_i; };
  const string& checksum(void) const { return checksum_i; };
  void checksum(const string& m) { checksum_i=m; checksum_b=true; };
  bool checksum_available(void) const { return checksum_b; };
  // TODO: rewrite 
  bool checksum_compare(const string& m) { return (checksum_i==m); };
  const struct tm* created(void) const { return &created_i; };
  void created(const struct tm* c) { memcpy(&created_i,c,sizeof(created_i)); created_b=true; };
  void created(const time_t* c);
  void created(const char* c);
  void created(const string& c) { created(c.c_str()); };
  bool created_available(void) const { return created_b; };
  bool created_compare(const char* c);
  // void md5_string(string &s);
  const list<string> & sources(void) const { return sources_; };
  void source(const string s) { sources_.push_back(s); };
  void source(const char* s) { sources_.push_back(string(s)); };
};

class SEState {
 friend ostream& operator<<(ostream& o,const SEState &s);
 private:
  int file_;
  int reg_;
  time_t file_last_changed_;
  time_t reg_last_changed_;
 public:
  SEState(void):file_(FILE_STATE_ACCEPTED),reg_(REG_STATE_LOCAL) {
    file_last_changed_=time(NULL);
    reg_last_changed_=time(NULL);
  };
  SEState(int f,int r):file_(f),reg_(r) {
    file_last_changed_=time(NULL);
    reg_last_changed_=time(NULL);
  };
  ~SEState() { };
  int file(void) const { return file_; };
  int reg(void) const { return reg_; };
  int file(int f) {
    if((f>=FILE_STATE_UNCHANGED) || (f<0)) return file_;
    file_last_changed_=time(NULL);
    return file_=f;
  };
  int reg(int r) {
    if((r>=REG_STATE_UNCHANGED) || (r<0)) return reg_;
    reg_last_changed_=time(NULL);
    return reg_=r;
  };
  bool set(const char* key,const char* value);
  time_t file_changed(void) { return file_last_changed_; };
  time_t reg_changed(void) { return reg_last_changed_; };
};

ostream& operator<<(ostream& o,const SEState &s);

class SEFileHandle;

// File itself
class SEFile: public SEAuthorization, public SEAttributes {
 private:
  bool valid;

  string path;          // Path to directory to store file
  const char* name;     // Name on file system

  SEFileRange* ranges;  // Ranges available (file is complete if NULL)
  LockSimple lock;      // use to atomically perform modifications (used 
                        // only from outside of that class)
  LockSimple lock_;     // Lock for internal operations
  int read_count;       // how many times it is open for reading
  int write_count;      // how many times it is open for writing
  int usage_count;      // not used yet
  int file_handle;      // Handle for file access
  time_t last_changed_;
  DiskSpaceLink space;  // Disk space request handler
  SEState state_;       // State of file and registration (is handled 
                        // outside of this class yet)

 public:
  operator bool(void) { return valid; };
  bool operator!(void) { return !valid; };
  /// Create new file
  SEFile(const char* dirpath,const SEAttributes&,DiskSpace& sp);
  /// Acquire stored file
  SEFile(const char *fname,DiskSpace& sp);
  ~SEFile(void);
  bool check_id(const char* id_) const { return (strcmp(id(),id_)==0); };
  bool check_name(const char* name_) const { return (strcmp(name,name_)==0); };
  int open(bool for_read);
  void close(bool for_read);
  uint64_t write(void* buf,uint64_t offset,uint64_t size);
  uint64_t read(void* buf,uint64_t offset,uint64_t size);
  int check_acl(const AuthUser &user);
  int write_acl(const AuthUser &user,const char* acl);
  int read_acl(const AuthUser &user,string& acl);
  int write_attr(void);
  bool state_file(int f);
  bool state_reg(int r);
  int state_file(void) { return state_.file(); };
  int state_reg(void) { return state_.reg(); };
  void acquire(void) { lock.block(); };     // lock for atomic operation
  void release(void) { lock.unblock(); };   // release after atomic operation
  void destroy(void);
  time_t state_file_changed(void) { return state_.file_changed(); };
  time_t state_reg_changed(void) { return state_.reg_changed(); };
  time_t data_changed(void) { return last_changed_; };
  int free_ranges(int num_ranges,SEFileRange ranges[]);
  bool full(void) { return (ranges==NULL); };
  int verify(void);
  int checksum_compute(const char* type);
};

#endif // __SE_FILE_H__
