#include "../../std.h"

#include <iostream>

#include "../../datamove/datapoint.h"
#include "../../datamove/datahandle.h"
#include "../../datamove/databufferpar.h"
#include "../../misc/log_time.h"

#include "se_replicator.h"


SEReplicator::SEReplicator(void) {

}

SEReplicator::~SEReplicator(void) {
  lock.block(); // after mutex destroyed waiting threads will be informed
}

// suppose this files has proper state - DOWNLOADING 
int SEReplicator::Obtain(SEFile& file) {
cerr<<"Obtain"<<endl;
  const list<string> &sources = file.sources();
  lock.block();
cerr<<"Obtain: size: "<<sources.size()<<endl;
  if(sources.size() == 0) { lock.unblock(); return -1; };
  bool downloaded = false;
  for(list<string>::const_iterator u = sources.begin();u!=sources.end();++u) {
    if(downloaded) break;
    // TODO - download from multiple sources simultaneously
    // Little hack to deal with wrong records created by fixed bug
    string u_ = *u;
    if(strncmp(u_.c_str(),"lrc://",6) == 0) u_.replace(0,3,"rls");
    cerr<<"Obtain: source: "<<u_<<endl;
    DataPoint source(u_.c_str());
    if(!source) {
      odlog(0)<<"Failed to acquire source: "<<source<<endl;
      continue;
    };
    if(!source.meta_resolve(true,UrlMap())) {
      odlog(0)<<"Failed to resolve source: "<<source<<endl;
      continue;
    };
    // Do some tricks with meta data (size and checksum so far).
    {
      bool modified = false;
      if(source.meta_size_available()) {
        unsigned long long int size = source.meta_size();
        if(!file.size_available()) file.size(size);
        modified=true;
      };
      if(source.meta_checksum_available()) {
        string check = source.meta_checksum();
        if(!file.checksum_available()) file.checksum(check);
        modified=true;
      };
      if(modified) file.write_attr();
    };
    if(!source.have_locations()) {
      odlog(0)<<"No locations for source found: "<<source<<endl;
      continue;
    };
    for(;;) {  // locations
      if(downloaded) break;
      if(!source.have_location()) {
        odlog(1)<<"Out of tries: "<<source<<endl;
        break;
      };
      DataHandle source_h(&source);
//      source_h.secure(force_secure);
//      source_h.passive(force_passive);
//      source_h.additional_checks(do_checks);
      int bufnum = 1;
      long int bufsize = (1024*1024);
      {
      DataHandle::analyze_t hint;
      source_h.analyze(hint);
      if(hint.bufnum>bufnum) bufnum=hint.bufnum;
      if(hint.bufsize>bufsize) bufsize=hint.bufsize;
      };
      bool seekable = source_h.out_of_order();
      if(bufnum>10) bufnum=10;
      if(bufsize>(10*1024*1024)) bufsize=10*1024*1024;
      bufnum=bufnum*2;
      odlog(2)<<"Creating buffer: "<<bufsize<<" x "<<bufnum<<endl;
//      CheckSum crc;
      DataBufferPar buffer(NULL,bufsize,bufnum);
      if(!buffer) {
        odlog(1)<<"Buffer creation failed !"<<endl;
        source.next_location();
        continue; // there is still hope to create buffer with different sizes 
      };
//      buffer.speed.set_min_speed(min_speed,min_speed_time);
//      buffer.speed.set_min_average_speed(min_average_speed);
//      buffer.speed.set_max_inactivity_time(max_inactivity_time);
//      buffer.speed.verbose(be_verbose);
//      buffer.speed.verbose(verbose_prefix);
      for(;;) { // ranges
        // choose range
        SEFileRange range;
        int n=file.free_ranges(1,&range);
        if(n==0) {
          odlog(0)<<"No ranges to download. File must be complete."<<endl;
          downloaded=true;
          break;
        };
        // TODO: decrease range if too big
        odlog(0)<<"Range to download: "<<range.start<<" - "<<range.end<<endl;
        if(file.open(false) != 0) { 
          odlog(0)<<"Failed to open file."<<endl;
          lock.unblock();
          return -1;
        };
        source_h.range(range.start,range.end);
        // download
        if(!source_h.start_reading(buffer)) {
          file.close(false);
          odlog(0)<<"Failed to start reading from source"<<endl;
          source.next_location();
          break;
        };
        unsigned long long int max_size = 0;
        for(;;) { // download
          /* 1. claim buffer */
          int h;
          unsigned int l;
          unsigned long long int p;
          if(!buffer.for_write(h,l,p,true)) {
            /* failed to get buffer - must be error or request to exit */
            if(!buffer.eof_read()) buffer.error_write(true);
            buffer.eof_write(true);
            break;
          };
          if(buffer.error()) {
            buffer.is_written(h);
            buffer.eof_write(true);
            break;
          };
          /* 2. write */
          if(file.write(buffer[h],p,l) != l) {
            odlog(0)<<"Failed to write to file."<<endl;
            buffer.is_written(h);
            buffer.error_write(true);
            buffer.eof_write(true);
            source_h.stop_reading();
            file.close(false);
            lock.unblock();
            return -1;
          };
          buffer.is_written(h);
          if((p+l) > max_size) max_size=p+l;
        }; // download
        source_h.stop_reading();
        if(buffer.error()) {
          file.close(false);
          odlog(0)<<"Error while downloading"<<endl;
          source.next_location();
          break;
        };
        if(!file.size_available()) {
          // That is not good and should not happen. But if it happened
          // try to fix that.
          if(source.meta_size_available()) {
            // 1. Try to use information obtained during resolving.
            file.size(source.meta_size()); file.write_attr();
            odlog(1)<<"Using size obtained from remote server: "<<file.size()<<" - "<<source.meta_size()<<endl;
          } else if(max_size <= range.end) { 
            // 2. Range was downloaded without error. If it is not
            // complete, that means file finished.
            file.size(max_size); file.write_attr();
            odlog(1)<<"Using size obtained from incomplete range: "<<file.size()<<" - "<<max_size<<endl;
          };
        };
        file.close(false);
      }; // ranges
    }; // location
  }; // sources
  lock.unblock();
  if(!downloaded) return -1;
  return 0;
}

