#include "../../std.h"

#include "../../misc/inttostring.h"
#include "../../misc/log_time.h"
#include "../../auth/auth.h"

#include "../httpsd.h"

HTTP_Service::HTTP_Service(void) {
}

HTTP_Service::~HTTP_Service(void) {
}

HTTP_Error HTTP_Service::get(const char* uri,int &keep_alive) {
  odlog(0)<<"HTTP_Service: virtual get - running incomplete service ?"<<endl;
  return HTTP_NOT_IMPLEMENTED;
}

HTTP_Error HTTP_Service::put(const char* uri,int &keep_alive) {
  odlog(0)<<"HTTP_Service: virtual put - running incomplete service ?"<<endl;
  return HTTP_NOT_IMPLEMENTED;
}

HTTP_Error HTTP_Service::post(const char* uri,int &keep_alive) {
  odlog(0)<<"HTTP_Service: virtual post - running incomplete service ?"<<endl;
  return HTTP_NOT_IMPLEMENTED;
}

static bool match_url(const string& uri,bool subtree,const char* u) {
  // 'u' contains full url (with trailing '/' stripped)
  // each service can be configured to serve either:
  // 1) path
  // 2) port+path
  // 3) full url
  const char* u_ = u;
  if((uri.length()==0) || (uri[0] == '/')) { // serving path
    u_=strchr(u_,':'); if(u_==NULL) return false; // bad url
    u_++; if((*u_) != '/') return false; // bad url
    u_++; if((*u_) != '/') return false; // bad url
    u_++; u_=strchr(u_,'/'); if(u_==NULL) u_=""; // bad url
  } else if(uri[0] == ':') {
    u_=strchr(u_,':'); if(u_==NULL) return false; // bad url
    u_++; if((*u_) != '/') return false; // bad url
    u_++; if((*u_) != '/') return false; // bad url
    u_++; u_=strchr(u_,':'); if(u_==NULL) return false; // bad url
  };
  if(subtree) {
    if(strncmp(u_,uri.c_str(),uri.length()) != 0) return false;
    if((u_[uri.length()] != '/') && (u_[uri.length()] != 0)) return false;
    return true;
  } else {
    return (strcmp(u_,uri.c_str()) == 0);
  };
}

bool HTTP_Service_description::match(const char* u) {
  return match_url(uri,properties.subtree,u);
}

bool HTTP_Service_item::match(const char *u) {
  return match_url(uri,subtree,u);
};


HTTP_Service* HTTP_Service_description::create(HTTPS_Connector& c) {
  return (*creator)(c,uri.c_str(),properties.arg);
}

HTTP_Service_description::HTTP_Service_description(const char* u,service_creator c,const HTTP_Service_Properties &prop):creator(c),uri(u),properties(prop) {
}


HTTP_Services::HTTP_Services(void) {
  pthread_mutex_init(&lock,NULL);
}

HTTP_Services::~HTTP_Services(void) {
  pthread_mutex_destroy(&lock);
}

bool HTTP_Services::add(const char* u,service_creator c,const HTTP_Service_Properties &prop) {
  pthread_mutex_lock(&lock);
  desc.push_back(HTTP_Service_description(u,c,prop));
  pthread_mutex_unlock(&lock);
  return true;
}

bool HTTP_Services::remove(const char* u,service_creator c) {
  pthread_mutex_lock(&lock);
  odlog(0)<<"IMPLEMET ME, PLEASE! Removing service '"<<u<<"'"<<endl;

/* !!!!!!!!!!!!!! TODO !!!!!!!!!!!!!!! */


  pthread_mutex_unlock(&lock);
  return true;
}

HTTP_Service* HTTP_Services::get(HTTPS_Connector& c,const char* u) {
  // 'u' contains full url
  pthread_mutex_lock(&lock);
  for(list<HTTP_Service_item>::iterator i = serv.begin();i!=serv.end();++i) {
    if((i->con == &c) && (i->match(u))) {
      odilog(2,c.pid)<<"Taking service from cache: "<<u<<endl;
      HTTP_Service* s = i->serv;
      pthread_mutex_unlock(&lock);
      return s;
    };
  };
  for(list<HTTP_Service_description>::iterator i = desc.begin();i!=desc.end();++i) {
    if(i->match(u)) {
      odilog(1,c.pid)<<"Creating new service: "<<u<<endl;
      HTTP_Service* s = i->create(c);
      if(s == NULL) break;
      serv.push_back(HTTP_Service_item(&c,i->uri.c_str(),i->properties.subtree,s)); 
      pthread_mutex_unlock(&lock);
      return s; 
    };
  };
  pthread_mutex_unlock(&lock);
  return NULL;
}

void HTTP_Services::close(const HTTPS_Connector& c) {
  pthread_mutex_lock(&lock);
  for(list<HTTP_Service_item>::iterator i = serv.begin();i!=serv.end();) {
    if(i->con == &c) {
      delete (i->serv);
      i=serv.erase(i);
    } else { ++i; };
  };
  pthread_mutex_unlock(&lock);
}


