#include "errors.h"
#include <string>

class ErrorContainer {
 friend class Error;
 private:
  string desc;
  int code;
  int subcode;
  pthread_mutex_t lock;
  int count;
  int inc(void) {
    pthread_mutex_lock(&lock);
    int res = ++count;
    pthread_mutex_unlock(&lock);
    return res;
  };
  int dec(void) {
    pthread_mutex_lock(&lock);
    int res = --count;
    pthread_mutex_unlock(&lock);
    return res;
  };
 public:
  ErrorContainer(int code_,int subcode_,const char *desc_):code(code_),subcode(subcode_),desc(desc_),count(0) {
    pthread_mutex_init(&lock,NULL);
  };
  ~ErrorContainer() { };
};

Error::Error(int code,int subcode,const char *desc) {
  err = new ErrorContainer(code,subcode,desc);
  err->inc();
}

Error::~Error() {
  if(err->dec() == 0) delete err;
}

Error::Error(int code,const char *desc) {
  Error(code,-1,desc);
}

Error::Error(const char *desc) {
  Error(-1,-1,desc);
}

void Error::operator=(const Error &e) {
  if(err->dec() == 0) delete err;
  err=e.err;
  err->inc();
}

int Error::get_code(void) {
  return err->code;
}

int Error::get_subcode(void) {
  return err->subcode;
}

const char* Error::get_desc(void) {
  return err->desc.c_str();
}

