/*
  this utility is used by globus-script-ng-submit to 'preparse'
  RSL and read some configuration from configuration file 
*/

#include "std.h"
#include <string>
#include <iostream>
#include <globus_common.h>
#include <globus_rsl.h>
#include "rsl/parse_rsl.h"
#include "config/config_pre.h"
#include "config/environment.h"
#include "misc/log_time.h"
#include "jobs/plugins.h"
#include "run/run_plugin.h"

int main(int argc,char** argv) {
  /* parse arguments */
  read_env_vars();
  int res=1;
  int n;
  char* filename = NULL;
  char* username = NULL;
  char* gramid   = NULL;
  while((n=getopt(argc,argv,"hf:u:i:")) != -1) {
    switch(n) {
      case ':': { cerr << "Missing argument" << endl; exit(1); };
      case 'f': {
        if((filename) || (username)) {
          cerr << "It should be only one argument" << endl; exit(1);
        };
        filename=optarg;
      }; break;
      case 'u': {
        username=optarg;
      }; break;
      case 'i': {
        gramid=optarg;
      }; break;
      case 'h': {
        cout<<"ng-parse-rsl -f rsl_file -u user_name -i job_id"<<endl;
      }; exit(1);
    };
  };
  if(!filename) { cerr << "Need filename" << endl; exit(1); };
  if(!username) { cerr << "Need username" << endl; exit(1); };
  if(globus_module_activate(GLOBUS_COMMON_MODULE) != GLOBUS_SUCCESS) {
    cerr<<"COMMON module activation failed\n"; res=1; exit(1);
  };
  if(globus_module_activate(GLOBUS_RSL_MODULE) != GLOBUS_SUCCESS) {
    cerr<<"RSL module activation failed\n"; res=1;
    globus_module_deactivate(GLOBUS_COMMON_MODULE); exit(1);
  };
  string action("");
  string jobid("");
  string control_dir("");
  string session_root("");
  string lrms("");
  string queue("");
  string default_lrms("");
  string default_queue("");
  ContinuationPlugins plugins;
  RunPlugin cred_plugin;
  LogTime::active=false;
  string allow_submit;
  if(!parse_rsl_for_action(filename,action,jobid,lrms,queue)) {
    cerr << "Failed parsing RSL" << endl; goto exit;
  };
  if(action == "request") {
    if(jobid.length() == 0) {
      if(!gramid) { cerr << "Need id" << endl; exit(1); };
      jobid=gramid;
    };
  };
  if((strcasecmp(action.c_str(),"request")) && 
     (strcasecmp(action.c_str(),"cancel")) && 
     (strcasecmp(action.c_str(),"clean"))) {
    cerr << "Unrecognizable action" << endl; goto exit;
  };
  if(!configure_user_dirs(username,control_dir,session_root,default_lrms,default_queue,plugins,cred_plugin,allow_submit)) {
    cerr << "Problem with configuration file" << endl; goto exit;
  };
  if(lrms.length()==0) lrms=default_lrms;
  if(queue.length()==0) queue=default_queue;
//  if(!preprocess_rsl(filename,session_root+"/"+jobid,jobid)) {
//    cerr << "Failed preprocessing RSL" << endl; goto exit;
//  };
  cout<<"'"<<action<<"' '"<<jobid<<"' '"<<
        control_dir<<"' '"<<session_root<<"' '"<<
        lrms<<"' '"<<queue<<"'"<<endl;
  res=0;
exit:
  globus_module_deactivate(GLOBUS_RSL_MODULE);
  globus_module_deactivate(GLOBUS_COMMON_MODULE);
  return 0;
}
