#include "../std.h"
#include <string>
#include <iostream>
#include <fstream>
#include <globus_common.h>
#include <globus_rsl.h>
#include <globus_gram_protocol.h>
#include "../jobs/users.h"
#include "../files/info_files.h"
#include "../files/info_types.h"
#include "../files/info_types.h"
#include "../transfer/replica_utils.h"
#include "../misc/canonical_dir.h"
#include "../misc/log_time.h"
#include "subst_rsl.h"
#include "parse_rsl.h"

void print_structure(globus_rsl_t *cur,int pad);
void print_value(globus_rsl_value_t* value);

int main(int argc,char* argv[]) {
  LogTime::active=false;
  if(argv[1] == NULL) return 1;
  string fname = argv[1];
  globus_rsl_t *rsl_tree = read_rsl(fname);
  if(rsl_tree == NULL) { 
    cerr << "Failed reading RSL" << endl; return 1;
  };
/*
  rsl_subst_table_t *symbol_table = NULL;
  symbol_table = (rsl_subst_table_t *) globus_libc_malloc
                            (sizeof(rsl_subst_table_t));
  rsl_subst_table_init(symbol_table);
  rsl_subst_table_insert(symbol_table,
                           strdup("NORDUGRID_LOCATION"),
                           strdup(DEFAULT_NORDUGRID_LOCATION));

  rsl_subst(rsl_tree,symbol_table);


  rsl_subst_table_destroy(symbol_table);
*/
  print_structure(rsl_tree,0);



//  if(globus_rsl_print_recursive(rsl_tree) != GLOBUS_SUCCESS) {
//    cerr << "Failed printing RSL" << endl;
//  };




  if(rsl_tree) globus_rsl_free_recursive(rsl_tree);
};


static char* rsl_operators [11] = {
  "?0?","=","!=",">",">=","<","<=","?7?","&","|","multi"
};

void print_structure(globus_rsl_t *cur,int pad) {
  for(int i=0;i<pad;i++) cout<<' ';
  if(globus_rsl_is_boolean(cur)) {
    cout<<"BOOLEAN: "<<cur->req.boolean.my_operator<<endl;
    globus_rsl_t* cur_;
    globus_list_t *list = cur->req.boolean.operand_list;
    for(;;) {
      if(globus_list_empty(list)) break;
      cur_=(globus_rsl_t*)globus_list_first(list);
      print_structure(cur_,pad+2);
      list=globus_list_rest(list);
    };
  }
  else if(globus_rsl_is_relation(cur)) {
    cout<<"RELATION: "<<cur->req.relation.attribute_name<<" ";
    cout<<rsl_operators[cur->req.relation.my_operator]<<" ";
    print_value(cur->req.relation.value_sequence);
    cout<<endl;
  }
  else {
    cout<<"UNKNOWN"<<endl;
  };
}

void print_value(globus_rsl_value_t* cur) {
  if(globus_rsl_value_is_literal(cur)) {
    cerr<<"LITERAL( ";
    cerr<<cur->value.literal.string<<" ";
  }
  else if(globus_rsl_value_is_variable(cur)) {
    cerr<<"VARIABLE( ";
    print_value(cur->value.variable.sequence);
  }
  else if(globus_rsl_value_is_concatenation(cur)) {
    cerr<<"CONCATENATION( ";
    print_value(cur->value.concatenation.left_value);
    print_value(cur->value.concatenation.right_value);
  }
  else if(globus_rsl_value_is_sequence(cur)) {
    cerr<<"SEQUENCE( ";
    globus_rsl_value_t* cur_;
    globus_list_t *list = cur->value.sequence.value_list;
    for(;;) {
      if(globus_list_empty(list)) break;
      cur_=(globus_rsl_value_t*)globus_list_first(list);
      print_value(cur_);
      list=globus_list_rest(list);
    };
  }
  else {
    cerr<<"UNKNOWN( ";
  };
  cerr<<") ";
}

typedef struct {
  char* name;
  char* value;
} value_pair_t;

/* 0 - false, 1 - true, -1 - not defined */
int eval_structure(globus_rsl_t *cur,globus_list_t *desc) {
  if(globus_rsl_is_boolean(cur)) {
    int is_and;
    if((is_and=globus_rsl_is_boolean_and(cur)) || 
       globus_rsl_is_boolean_or(cur)) {
      int had_undef=0;
      int cur_value=1;
      /* go through list of operands */
      globus_rsl_t* cur_;
      globus_list_t *list = cur->req.boolean.operand_list;
      for(;;) {
        if(globus_list_empty(list)) break;
        cur_=(globus_rsl_t*)globus_list_first(list);
        cur_value=eval_structure(cur_,desc);
        if(is_and) {
          if(cur_value == 0) return 0;
          if(cur_value == -1) had_undef=1;
        }
        else {
          if(cur_value == 1) return 1;
          if(cur_value == -1) had_undef=1;
        };
        list=globus_list_rest(list);
      };
      if(had_undef) return -1;
      return 1;
    }
    else {
      return -1; /* unsupported boolean operator */
    };
  }
  else if(globus_rsl_is_relation(cur)) {
    /* find values */
    value_pair_t* cur_val;
    for(;;) {
      if(globus_list_empty(desc)) break;
      cur_val=(value_pair_t*)globus_list_first(desc);
      if(!strcmp(cur_val->name,cur->req.relation.attribute_name)) {
        /* calculate here */
        globus_rsl_value_t* values = cur->req.relation.value_sequence;
        if(globus_rsl_value_is_literal(values)) {
          switch(cur->req.relation.my_operator) {
            case GLOBUS_RSL_EQ: {

            }; break;
            case GLOBUS_RSL_NEQ: {

            }; break;
            case GLOBUS_RSL_GT: {

            }; break;
            case GLOBUS_RSL_GTEQ: {

            }; break;
            case GLOBUS_RSL_LT: {

            }; break;
            case GLOBUS_RSL_LTEQ: {

            }; break;
            default: return -1;
          };
        }
        else { return -1; };
      };
      desc=globus_list_rest(desc);
    };
    return -1;
  }
  else {
    return -1;  /* this can't be evaluated */
  };
}
