#!/bin/bash

# Benchmark difference between normal ssh and gridssh

ITERATIONS="10"
HOST="tmp.imada.sdu.dk"
PORT="2222"
USER="grid"

if [ $# -eq 4 ]; then
    ITERATIONS="$1"
    HOST="$2"
    PORT="$3"
    USER="$4"
elif [ $# -eq 3 ]; then
    ITERATIONS="$1"
    HOST="$2"
    PORT="$3"
elif [ $# -eq 2 ]; then
    ITERATIONS="$1"
    HOST="$2"
elif [ $# -eq 1 ]; then
    ITERATIONS="$1"
else
    echo using defaults
fi

echo "***************************************************************************"
echo "* This benchmark requires ssh-agent passphrase handling to make any sense *"
echo "***************************************************************************"

function benchmark_gridssh() {

	for i in `seq 1 $ITERATIONS`; do
		gridssh -p $PORT $HOST pwd > /dev/null;
	done
}

function benchmark_gridssh_cached() {

	for i in `seq 1 $ITERATIONS`; do
		gridssh -C -p $PORT $HOST pwd > /dev/null;
	done
}

function benchmark_ssh() {

	for i in `seq 1 $ITERATIONS`; do
		ssh -l$USER $HOST pwd > /dev/null;
	done
}

echo "Benchmarking $ITERATIONS iterations against $HOST:$PORT"

# First time gridssh

time benchmark_gridssh

time benchmark_gridssh_cached

time benchmark_ssh

exit 0
