/*
  
   Copyright 2004 Jonas Bardino

   This file is part of NorduGRID GridSSH.

    NorduGRID GridSSH is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    NorduGRID GridSSH is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with NorduGRID GridSSH; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifndef __GSSHD_CONF_H__
#define __GSSHD_CONF_H__

#define GRIDSSHD_VERSION "0.1.4"

/* Include headers for all used globus module for completeness even though 
   some of them would actually be implicitly included in any case */
/* Please note that changes to the inclusions may have to be reflected in 
   key_handling.h */
#include <globus_common.h>
#include <globus_openssl.h>
#include <globus_gss_assist.h>
#include <globus_gsi_credential.h>
#include <globus_io.h>

#include "expire_queue.h"

/* Server defaults */
#define DEFAULT_MAX_PROCS 64 
#define DEFAULT_MAX_CONNECTIONS 32
#define DEFAULT_BIND_RETRIES 5
#define DEFAULT_PID_FILE "/var/run/gridsshd.pid"

#define DEFAULT_SLEEP_TIME 10000
#define DEFAULT_BIND_SLEEP 100000

#define MAX_PORT 65536

/* Server errors */
#define MODULE_ACTIVATION_ERR 1
#define SERVER_CRED_ERR 10
#define SOCKET_CREATE_ERR 20
#define SOCKET_BIND_ERR 21
#define SOCKET_LISTEN_ERR 22
#define CLIENT_AUTH_FAILURE 30

/* Environment */
#define DEFAULT_GRIDMAP_FILE "/etc/grid-security/grid-mapfile"
#define DEFAULT_OUT_FILE "/var/log/gridsshd.log"
#define DEFAULT_ERR_FILE "/var/log/gridsshd.log"
#define DEFAULT_AUTH_FILENAME ".ssh/authorized_keys"
#define DEFAULT_EXPIRE_MODE PROXY_EXPIRE
/* Run expiry check every X seconds (5 minutes here) */
#define DEFAULT_EXPIRE_INTERVAL 5*60

/* Prototypes */
int init_conf(int argc, char **argv);
int del_conf(void);

int initialize_threads(globus_mutex_t *threads_mutex);
int finalize_threads(globus_mutex_t *threads_mutex);

int inc_active_threads(globus_mutex_t *threads_mutex);
int dec_active_threads(globus_mutex_t *threads_mutex);

void sig_handler(int sig);

int finalize(int level);

int globus_activation(void);
int globus_deactivation(void);
void handle_session(globus_io_handle_t *conn_handle);
void listen_callback(void *callback_arg, globus_io_handle_t *conn_handle, globus_result_t result);
globus_bool_t auth_callback(void *arg, globus_io_handle_t *handle, globus_result_t res, char *identity, gss_ctx_id_t context);
char *extract_openssh_pubkey(char *deleg_proxy_filename);
void print_usage(void);

#endif
