/*
  
   Copyright 2004 Jonas Bardino

   This file is part of NorduGRID GridSSH.

    NorduGRID GridSSH is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    NorduGRID GridSSH is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with NorduGRID GridSSH; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*
  This file is a derivative of the original OpenSSH-3.8p1 
  (http://www.openssh.org) key-handling code. It was developed by 
  Jonas Bardino for the GridSSH subproject of the NorduGRID ARC 
  project (http://www.nordugrid.org).
  
  The purpose of this particular code is to translate a EVP_PKEY extracted 
  from a x509 certificate to the format used by OpenSSH.
  
  The files key_handling.[ch] include bits and pieces from the following 
  original OpenSSH files:
  key.[ch], buffer.[ch], bufaux.[ch] and uuencode.[ch]
  The original copyright notices from each of those files follow here:

key.c
* read_bignum():
* Copyright (c) 1995 Tatu Ylonen <ylo@cs.hut.fi>, Espoo, Finland
*
* As far as I am concerned, the code I have written for this software
* can be used freely for any purpose.  Any derived versions of this
* software must be clearly marked as such, and if the derived work is
* incompatible with the protocol description in the RFC file, it must be
* called by a name other than "ssh" or "Secure Shell".
*
*
* Copyright (c) 2000, 2001 Markus Friedl.  All rights reserved.
*
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
* 1. Redistributions of source code must retain the above copyright
*    notice, this list of conditions and the following disclaimer.
* 2. Redistributions in binary form must reproduce the above copyright
*    notice, this list of conditions and the following disclaimer in the
*    documentation and/or other materials provided with the distribution.
*
* THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
* IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
* OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
* IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
* INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
* DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
* THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
* THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*

key.h
* Copyright (c) 2000, 2001 Markus Friedl.  All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
* 1. Redistributions of source code must retain the above copyright
*    notice, this list of conditions and the following disclaimer.
* 2. Redistributions in binary form must reproduce the above copyright
*    notice, this list of conditions and the following disclaimer in the
*    documentation and/or other materials provided with the distribution.
*
* THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
* IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
* OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
* IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
* INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
* DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
* THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
* THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


buffer.c
* Author: Tatu Ylonen <ylo@cs.hut.fi>
* Copyright (c) 1995 Tatu Ylonen <ylo@cs.hut.fi>, Espoo, Finland
*                    All rights reserved
* Functions for manipulating fifo buffers (that can grow if needed).
*
* As far as I am concerned, the code I have written for this software
* can be used freely for any purpose.  Any derived versions of this
* software must be clearly marked as such, and if the derived work is
* incompatible with the protocol description in the RFC file, it must be
* called by a name other than "ssh" or "Secure Shell".

buffer.h
* Author: Tatu Ylonen <ylo@cs.hut.fi>
* Copyright (c) 1995 Tatu Ylonen <ylo@cs.hut.fi>, Espoo, Finland
*                    All rights reserved
* Code for manipulating FIFO buffers.
*
* As far as I am concerned, the code I have written for this software
* can be used freely for any purpose.  Any derived versions of this
* software must be clearly marked as such, and if the derived work is
* incompatible with the protocol description in the RFC file, it must be
* called by a name other than "ssh" or "Secure Shell".


bufaux.c
* Author: Tatu Ylonen <ylo@cs.hut.fi>
* Copyright (c) 1995 Tatu Ylonen <ylo@cs.hut.fi>, Espoo, Finland
*                    All rights reserved
* Auxiliary functions for storing and retrieving various data types to/from
* Buffers.
*
* As far as I am concerned, the code I have written for this software
* can be used freely for any purpose.  Any derived versions of this
* software must be clearly marked as such, and if the derived work is
* incompatible with the protocol description in the RFC file, it must be
* called by a name other than "ssh" or "Secure Shell".
*
*
* SSH2 packet format added by Markus Friedl
* Copyright (c) 2000 Markus Friedl.  All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
* 1. Redistributions of source code must retain the above copyright
*    notice, this list of conditions and the following disclaimer.
* 2. Redistributions in binary form must reproduce the above copyright
*    notice, this list of conditions and the following disclaimer in the
*    documentation and/or other materials provided with the distribution.
*
* THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
* IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
* OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
* IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
* INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
* DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
* THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
* THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

bufaux.h
* Author: Tatu Ylonen <ylo@cs.hut.fi>
* Copyright (c) 1995 Tatu Ylonen <ylo@cs.hut.fi>, Espoo, Finland
*                    All rights reserved
*
* As far as I am concerned, the code I have written for this software
* can be used freely for any purpose.  Any derived versions of this
* software must be clearly marked as such, and if the derived work is
* incompatible with the protocol description in the RFC file, it must be
* called by a name other than "ssh" or "Secure Shell".


uuencode.c
* Copyright (c) 2000 Markus Friedl.  All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
* 1. Redistributions of source code must retain the above copyright
*    notice, this list of conditions and the following disclaimer.
* 2. Redistributions in binary form must reproduce the above copyright
*    notice, this list of conditions and the following disclaimer in the
*    documentation and/or other materials provided with the distribution.
*
* THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
* IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
* OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
* IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
* INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
* DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
* THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
* THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

uuencode.h
* Copyright (c) 2000 Markus Friedl.  All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
* 1. Redistributions of source code must retain the above copyright
*    notice, this list of conditions and the following disclaimer.
* 2. Redistributions in binary form must reproduce the above copyright
*    notice, this list of conditions and the following disclaimer in the
*    documentation and/or other materials provided with the distribution.
*
* THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
* IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
* OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
* IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
* INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
* DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
* THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
* THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#ifndef __KEY_HANDLING_H__
#define __KEY_HANDLING_H__

/*
 *
 * Key handling derived from OpenSSH key.c code
 *
 */

/* Changed to use output functions from log.c in general instead of 
   original ones */

/* Changed to include some globus modules (which supplies openssl functions 
   implicitly) in .h file rather than the original openssl headers. */
/* #include <openssl/evp.h> */

#include <stdlib.h> /* malloc */
#include <stdio.h> /* fopen, etc */
#include <string.h> /* memmove */
#include <resolv.h> /* uuencode */
#include "key_handling.h"
#include "log.h"


/* Modified version of key_new that inits a key struct from an existing key */
Key *key_init(int type, void *key)
{
  /* Changed to include more robustness: always init pointers to NULL and check 
     malloc. */
	Key *k = NULL;
	if((k = malloc(sizeof(*k))) == NULL) return NULL;
	k->type = type;
	k->flags = 0;
	k->dsa = NULL;
	k->rsa = NULL;
	switch (k->type) {
	case KEY_RSA1:
	case KEY_RSA:
	    k->rsa = (RSA *)key;
	    break;
	case KEY_DSA:
	    k->dsa = (DSA *)key;
	    break;
	case KEY_UNSPEC:
		break;
	default:
		errmsg("key_new: bad key type %d", k->type);
		return NULL;
		break;
	}
	return k;
}

void key_free(Key *k)
{
  /* Changed to not free rsa/dsa entry as it should be freed from caller of 
     key_init */
	switch (k->type) {
	case KEY_RSA1:
	case KEY_RSA:
/* 		if (k->rsa != NULL) */
/* 			RSA_free(k->rsa); */
		k->rsa = NULL;
		break;
	case KEY_DSA:
/* 		if (k->dsa != NULL) */
/* 			DSA_free(k->dsa); */
		k->dsa = NULL;
		break;
	case KEY_UNSPEC:
		break;
	default:
		errmsg("key_free: bad key type %d", k->type);
		return;
		break;
	}
	free(k);
}

/* Modified version of write_bignum that writes to a string instead of a file */
static int write_bignum_str(char *str, BIGNUM *num)
{
	char *buf = BN_bn2dec(num);
	if (buf == NULL) {
		errmsg("write_bignum: BN_bn2dec() failed");
		return 0;
	}
	sprintf(str, " %s", buf);
	OPENSSL_free(buf);
	return 1;
}

/* Modified key_write that writes to string instead of file */
int key_write_str(const Key *key, char **key_str)
{
  /* Changed to include more robustness: always init pointers to NULL and check 
     malloc. */
	int n, success = 0, bit_chars = 1;
	u_int len, bits = 0;
	u_char *blob = NULL;
	char *uu = NULL;
	
	if(!key_str) {
	  errmsg("passed NULL string pointer to %s",__FUNCTION__);
	  return success;
	}
	if(!key) {
	  errmsg("passed NULL key pointer to %s",__FUNCTION__);
	  return success;
	}

	if (key->type == KEY_RSA1 && key->rsa != NULL) {
	  /* size of modulus 'n' */
	  bits = BN_num_bits(key->rsa->n);
	  n = bits;
	  while((n/=10) > 0) bit_chars++;
	  if(!(*key_str) && !(*key_str = malloc((bit_chars+2+strlen(BN_bn2dec(key->rsa->n))+strlen(BN_bn2dec(key->rsa->e))+1)*sizeof(char)))) {
	    errmsg("malloc() failed to allocate openssh_pubkey string");
	    return success;
	  }
	  sprintf(*key_str, "%u", bits);
	  if (write_bignum_str(*key_str, key->rsa->e) &&
	      write_bignum_str(*key_str, key->rsa->n)) {
	    success = 1;
	  } else {
	    errmsg("key_write: failed for RSA key");
	  }
	} else if ((key->type == KEY_DSA && key->dsa != NULL) ||
	    (key->type == KEY_RSA && key->rsa != NULL)) {
		key_to_blob(key, &blob, &len);
		uu = malloc(2*len);
		n = uuencode(blob, len, uu, 2*len);
		if (n > 0) {
		  if(!(*key_str) && !(*key_str = malloc((strlen(key_ssh_name(key)) + 1 + 2*len + 1)*sizeof(char)))) {
		    errmsg("malloc() failed to allocate openssh_pubkey string");
		    return success;
		  }

		  sprintf(*key_str, "%s %s", key_ssh_name(key), uu);
		  success = 1;
		}
		free(blob);
		free(uu);
	}
	return success;
}

const char *key_ssh_name(const Key *k)
{
	switch (k->type) {
	case KEY_RSA:
		return "ssh-rsa";
		break;
	case KEY_DSA:
		return "ssh-dss";
		break;
	}
	return "ssh-unknown";
}

int key_type_from_name(char *name)
{
	if (strcmp(name, "rsa1") == 0) {
		return KEY_RSA1;
	} else if (strcmp(name, "rsa") == 0) {
		return KEY_RSA;
	} else if (strcmp(name, "dsa") == 0) {
		return KEY_DSA;
	} else if (strcmp(name, "ssh-rsa") == 0) {
		return KEY_RSA;
	} else if (strcmp(name, "ssh-dss") == 0) {
		return KEY_DSA;
	}
	warnmsg(0,"key_type_from_name: unknown key type '%s'", name);
	return KEY_UNSPEC;
}


int key_to_blob(const Key *key, u_char **blobp, u_int *lenp)
{
	Buffer b;
	int len;

	if (key == NULL) {
	  errmsg("key_to_blob: key == NULL");
	  return 0;
	}
	buffer_init(&b);
	switch (key->type) {
	case KEY_DSA:
		buffer_put_cstring(&b, key_ssh_name(key));
		buffer_put_bignum2(&b, key->dsa->p);
		buffer_put_bignum2(&b, key->dsa->q);
		buffer_put_bignum2(&b, key->dsa->g);
		buffer_put_bignum2(&b, key->dsa->pub_key);
		break;
	case KEY_RSA:
		buffer_put_cstring(&b, key_ssh_name(key));
		buffer_put_bignum2(&b, key->rsa->e);
		buffer_put_bignum2(&b, key->rsa->n);
		break;
	default:
		errmsg("key_to_blob: unsupported key type %d", key->type);
		buffer_free(&b);
		return 0;
	}
	len = buffer_len(&b);
	if (lenp != NULL)
		*lenp = len;
	if (blobp != NULL) {
		*blobp = malloc(len);
		memcpy(*blobp, buffer_ptr(&b), len);
	}
	memset(buffer_ptr(&b), 0, len);
	buffer_free(&b);
	return len;
}

/*
 *
 * Buffer manipulation derived from OpenSSH buffer.c and bufaux.c code
 *
 */

void buffer_init(Buffer *buffer)
{
	const u_int len = 4096;

	buffer->alloc = 0;
	buffer->buf = malloc(len);
	buffer->alloc = len;
	buffer->offset = 0;
	buffer->end = 0;
}

/* Frees any memory used for the buffer. */

void buffer_free(Buffer *buffer)
{
	if (buffer->alloc > 0) {
		memset(buffer->buf, 0, buffer->alloc);
		buffer->alloc = 0;
		free(buffer->buf);
	}
}

/* Appends data to the buffer, expanding it if necessary. */

void buffer_append(Buffer *buffer, const void *data, u_int len)
{
	void *p;
	p = buffer_append_space(buffer, len);
	memcpy(p, data, len);
}

/*
 * Appends space to the buffer, expanding the buffer if necessary. This does
 * not actually copy the data into the buffer, but instead returns a pointer
 * to the allocated region.
 */

void *buffer_append_space(Buffer *buffer, u_int len)
{
	u_int newlen;
	void *p = NULL;

	if (len > 0x100000) {
		errmsg("buffer_append_space: len %u not supported", len);
		return p;
	}

	/* If the buffer is empty, start using it from the beginning. */
	if (buffer->offset == buffer->end) {
		buffer->offset = 0;
		buffer->end = 0;
	}
restart:
	/* If there is enough space to store all data, store it now. */
	if (buffer->end + len < buffer->alloc) {
		p = buffer->buf + buffer->end;
		buffer->end += len;
		return p;
	}
	/*
	 * If the buffer is quite empty, but all data is at the end, move the
	 * data to the beginning and retry.
	 */
	if (buffer->offset > buffer->alloc / 2) {
		memmove(buffer->buf, buffer->buf + buffer->offset,
			buffer->end - buffer->offset);
		buffer->end -= buffer->offset;
		buffer->offset = 0;
		goto restart;
	}
	/* Increase the size of the buffer and retry. */

	newlen = buffer->alloc + len + 32768;
	if (newlen > 0xa00000) {
		errmsg("buffer_append_space: alloc %u not supported",
		    newlen);
		return NULL;
	}

	buffer->buf = realloc(buffer->buf, newlen);
	buffer->alloc = newlen;
	goto restart;
	/* NOTREACHED */
}

/* Returns the number of bytes of data in the buffer. */

u_int buffer_len(Buffer *buffer)
{
	return buffer->end - buffer->offset;
}

/* Returns a pointer to the first used byte in the buffer. */

void *buffer_ptr(Buffer *buffer)
{
	return buffer->buf + buffer->offset;
}


/*
 * Stores an BIGNUM in the buffer in SSH2 format.
 */
void buffer_put_bignum2(Buffer *buffer, const BIGNUM *value)
{
  /* 
     Changed 'u_int bytes' to 'int bytes'.
     Openssl docs specify that BN_num_bytes returns an int and we don't want
     complaints from the compiler about unsigned to signed int comparison 
     (oi != bytes-1) .
     Added pointer initialization to NULL.

     -Jonas
  */
	int bytes;
	u_char *buf = NULL;
	int oi;
	u_int hasnohigh = 0;

	if (BN_is_zero(value)) {
		buffer_put_int(buffer, 0);
		return;
	}
	if (value->neg) {
	  errmsg("buffer_put_bignum2: negative numbers not supported");
	  return;
	}
	bytes = BN_num_bytes(value) + 1; /* extra padding byte */
	if (bytes < 2) {
	  errmsg("buffer_put_bignum2: BN too small");
	  return;
	}
	buf = malloc(bytes);
	buf[0] = '\0';
	/* Get the value of in binary */
	oi = BN_bn2bin(value, buf+1);
	if (oi != bytes-1) {
	  errmsg("buffer_put_bignum2: BN_bn2bin() failed: ");
	  errmsg("oi %d != bin_size %d", oi, bytes);
	  return;
	}
	hasnohigh = (buf[1] & 0x80) ? 0 : 1;
	buffer_put_string(buffer, buf+hasnohigh, bytes-hasnohigh);
	memset(buf, 0, bytes);
	free(buf);
}

void buffer_put_int(Buffer *buffer, u_int value)
{
	char buf[4];

	PUT_32BIT(buf, value);
	buffer_append(buffer, buf, 4);
}

/*
 * Stores and arbitrary binary string in the buffer.
 */
void buffer_put_string(Buffer *buffer, const void *buf, u_int len)
{
	buffer_put_int(buffer, len);
	buffer_append(buffer, buf, len);
}
void buffer_put_cstring(Buffer *buffer, const char *s)
{
  if (s == NULL) {
    errmsg("buffer_put_cstring: s == NULL");
    return;
  }
  buffer_put_string(buffer, s, strlen(s));
}

/*
 *
 * Uuencoding derived from OpenSSH uuencode.c code
 *
 */

int uuencode(const u_char *src, u_int srclength, char *target, size_t targsize)
{
  return __b64_ntop(src, srclength, target, targsize);
}

#endif
