/*
  
   Copyright 2004 Jonas Bardino

   This file is part of NorduGRID GridSSH.

    NorduGRID GridSSH is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    NorduGRID GridSSH is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with NorduGRID GridSSH; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/* 
   This header file is a derivative of a number of similar header files from 
   the OpenSSH project. Please refer to key_handling.c for a full list 
   including the original copyright notices
*/

#ifndef KEY_HANDLING_H
#define KEY_HANDLING_H

/*
 *
 * Key handling derived from OpenSSH key.h code
 *
 */

/* Changed to include globus_openssl and globus_gsi_credential 
   (which supplies openssl functions implicitly) rather than openssl 
   headers here and in .c file.
   -Jonas
*/
/* #include <openssl/rsa.h> */
/* #include <openssl/dsa.h> */
/* #include <openssl/bn.h> */
#include <globus_openssl.h>
#include <globus_gsi_credential.h>

typedef struct Key Key;
enum types {
	KEY_RSA1,
	KEY_RSA,
	KEY_DSA,
	KEY_UNSPEC
};

/* Macro derived from OpenSSH getput.h for storing data in msb first order. */

#define PUT_32BIT(cp, value) ((cp)[0] = (value) >> 24, \
			      (cp)[1] = (value) >> 16, \
			      (cp)[2] = (value) >> 8, \
			      (cp)[3] = (value))

struct Key {
	int	 type;
	int	 flags;
	RSA	*rsa;
	DSA	*dsa;
};


Key             *key_init(int type, void *key);
void		 key_free(Key *);
int		 key_write_str(const Key *, char **);
int		 key_to_blob(const Key *, u_char **, u_int *);
const char	*key_ssh_name(const Key *);


/*
 *
 * Buffer manipulation derived from OpenSSH buffer.h and bufaux.h code
 *
 */

typedef struct {
  u_char  *buf;           /* Buffer for data. */
  u_int    alloc;         /* Number of bytes allocated for data. */
  u_int    offset;        /* Offset of first byte containing data. */
  u_int    end;           /* Offset of last byte containing data. */
}       Buffer;

void     buffer_init(Buffer *);
void     buffer_free(Buffer *);

u_int    buffer_len(Buffer *);
void    *buffer_ptr(Buffer *);

void	 buffer_append(Buffer *, const void *, u_int);
void	*buffer_append_space(Buffer *, u_int);

void    buffer_put_bignum2(Buffer *, const BIGNUM *);
void    buffer_put_int(Buffer *, u_int);
void    buffer_put_string(Buffer *, const void *, u_int);
void    buffer_put_cstring(Buffer *, const char *);

/*
 *
 * Uuencoding derived from OpenSSH uuencode.h code
 *
 */

int uuencode(const u_char *, u_int, char *, size_t);

#endif

