#include <vector>
#include <string>
#include <fstream>
#include <iostream>
#include <cstdlib>

#include "JobStatWindow.h"

#include "qlistview.h"
#include "qlayout.h"
#include "qpushbutton.h"
#include "qlabel.h"

using namespace std;

JobStatWindow::JobStatWindow( QWidget* parent, const char* name ) : QWidget( parent, name )
{ 
  listView = new QListView( this );
  listView->addColumn( "JobIDs" );
  listView->addColumn( "JobStatus" );
  listView->addColumn( "JobNames" );
  listView->setColumnWidthMode( 1, QListView::Maximum );
  //  listView->setResizeMode( QListView::LastColumn );
  listView->setSelectionMode( QListView::Multi );

  invertJobsButton = new QPushButton( "Invert selected jobs", this );
  connect( invertJobsButton, SIGNAL( pressed( ) ), listView, SLOT( invertSelection( ) ) );

  allJobsButton = new QPushButton( "Select all jobs", this );
  connect( allJobsButton, SIGNAL( pressed( ) ), this, SLOT( selectAllJobs( ) ) );

  updateJobsButton = new QPushButton( "Update selected jobs", this );
  connect( updateJobsButton, SIGNAL( pressed( ) ), this, SLOT( updateJobs( ) ) );

  renewJobsButton = new QPushButton( "Renew proxy for selected jobs", this );
  connect( renewJobsButton, SIGNAL( pressed( ) ), this, SLOT( renewJobs( ) ) );

  syncJobsButton = new QPushButton( "Synchronize joblist with IS", this );
  connect( syncJobsButton, SIGNAL( pressed( ) ), this, SLOT( syncJobs( ) ) );

  jobLabel = new QLabel( "Job window", this );

  screenLayout = new QGridLayout( this, 4, 3, 10 );
  screenLayout->addMultiCellWidget( listView, 1, 1, 0, 2 );
  screenLayout->addWidget( invertJobsButton, 2, 0 );
  screenLayout->addWidget( allJobsButton, 3, 0 );
  screenLayout->addWidget( renewJobsButton, 2, 2 );
  screenLayout->addWidget( syncJobsButton, 3, 2 );
  screenLayout->addWidget( updateJobsButton, 2, 1 );
  screenLayout->addWidget( jobLabel, 0, 0 );
};

JobStatWindow::~JobStatWindow()
{ 
  delete screenLayout;
  delete listView;
  delete updateJobsButton;
  delete invertJobsButton;
  delete syncJobsButton;
  delete renewJobsButton;
};

void JobStatWindow::updateTab( )
{
  readJobs();
};

void JobStatWindow::readJobs( )
{
  string filename = getenv( "HOME" );
  filename += "/.ngjobs";

  ifstream jobfile( filename.c_str() );
  if( !jobfile ) {
    emit linemessage( "Cannot open job-file ..." );
    return;
  }
  
  string line;
  QListViewItem* item;

  while( getline( jobfile, line ) ) {
    unsigned int mark = line.find('#');
    if( mark == string::npos ) continue;
    
    string jobid = line.substr( 0, mark-1 );
    string jobname = line.substr( mark+1 );

    item = new QListViewItem( listView, jobid.c_str(), "UNKNOWN", jobname.c_str() );
  }

  jobfile.close();
  return;
}

void JobStatWindow::selectAllJobs( )
{
  listView->selectAll( true );
}

void JobStatWindow::updateJobs( )
{
}

void JobStatWindow::renewJobs( )
{
}

void JobStatWindow::syncJobs( )
{
}
