#!/usr/bin/perl
use lib "$ENV{PWD}/";
use lib "$ENV{NORDUGRID_LOCATION}/libexec";
use lib "/opt/nordugrid/libexec";
use infosys_shared;
use InfosysQJU qw(%config %hoh_gmjobs %users &queueldif &jobsldif &usersldif);
#use warnings;


# generates everything under the nordugrid-queue-name=xz,
# creates the queue, job & user entries.
# Real GM and config values are used, but dummy values
# are set for LRMS-dependent  attributes (13, 66 or "dummy").
# The GM interface is factored into the InfosysQJU module.
# Use this template to port the infosys.
# Don't use 'my xyz' for the interfacing GLOBAL variables!


#remove/unset this variable for a real script
$DUMMY_LRMS_VALUES = "true";

$timestamp = time;


#<LRMSPORT   
# read the queue info such as limits, authorized users' list, etc from the LRMS "queue" command
# into %queue_info and  @acl_users (list of LRMS-authorized unix users, undefined list means
# every unix user is LRMS authorized)

# queue limits, setting dummy values
# don't set the values if they are not relevant to the LRMS
# nordugrid-queue-maxrunning
$queue_info{"max_running"} = 13;
# nordugrid-queue-maxinqueue
# dummy value
$queue_info{"max_queuable"} = 133;
# nordugrid-queue-maxuserrun
# dummy value
$queue_info{"max_user_run"} = 13;
# nordugrid-queue-maxcputime (in minutes)
# dummy value
$queue_info{"max.cput"} = 133;
# nordugrid-queue-mincput (in minutes)
# dummy value
$queue_info{"min.cput"} = 0;
# nordugrid-queue-defaultcputime (in minutes)
# dummy value
$queue_info{"default.cput"} = 133;

#LRMSPORT>



#<LRMSPORT
# read the job info from LRMS to a hash of hashes %hoh_lrmsjobs
# The LRMS jobid is the key of the hash, it MUST match the one stored
# in the GM's job.xx.local file as 'localid'
# determine the queue rank of a LRMS queuing job
# $hoh_lrmsjobs{$jobid}{"rank"} = 1 
# example dummy values for a dummy job with the 'dummy13' LRMS jobid:

#dummy value for LRMS state
$hoh_lrmsjobs{"dummy13"}{"job_state"} = "DUMMY_LRMS_STATE";   
# dummy value for queue position
$hoh_lrmsjobs{"dummy13"}{"rank"} = 13;
# dummy value for used memory				     	      
$hoh_lrmsjobs{"dummy13"}{"used.mem"} = "13";	      
# dummy value (in minutes) for used walltime
$hoh_lrmsjobs{"dummy13"}{"used.walltime"} = 13;
# dummy value (in minutes) for used cputime
$hoh_lrmsjobs{"dummy13"}{"used.cput"} = 13;		
# dummy value (in minutes) for requested cputime
$hoh_lrmsjobs{"dummy13"}{"req.cput"} = "66";
#nordugrid-job-executionnodes, notice the 'syntax' of multiple nodes
$hoh_lrmsjobs{"dummy13"}{"exec_host"} = "dummy1+dummy2+dummy3";
# comment string from the LRMS
$hoh_lrmsjobs{"dummy13"}{"comment"} = "Running/queuing in dummy LRMS"; 		

#LRMSPORT>


#<LRMSPORT
# scan the LRMS and count the number of running jobs with their multiciplity
# and the number or queued jobs 
# count the jobs per unix users and the number of queueing jobs per users
# These values may be needed for the calculation of the per user freecpus
# useful variables: $user_jobs_queued{$username}, $user_jobs_running{$username}
# required values for the queue entry: $totalrunning_in_the_queue, $totalqueued
#dummy values
# nordugrid-queue-running
$totalrunning_in_the_queue = 13;
# nordugrid-queue-queued
$totalqueued = 13;
#LRMSPORT>



#<LRMSPORT
# scan the LRMS in order to find out the $totalcpus and the $usedcpus
# Take into account the $dedicated_node_string or the $queue_node_string
# for distinguishing grid and non-grid enabled nodes (skip nonededicated nodes).
# The $totalcpus and the $usedcpus may be needed for the calculation of the per user freecpus.
#LRMSPORT>



if ($config{loglevel} == 2) {  
   my $runtime = time - $timestamp; 
   &infosys_shared::write_log("LRMS I/O time: $runtime");  
}


#<LRMSPORT
# Count the running and queued GRID jobs in the LRMS (a subset of all LRMS jobs)
# running jobs are counted with their multiciplity
# you can loop through the hoh_gmjobs and use the "dummy13"=$hoh_gmjobs{$ID}{"localid"}
# and the $hoh_lrmsjobs{"dummy13"}{"job_state"} to find queuing and running grid jobs
# in the LRMS
# Provide the $gridrunning, $gridqueued numbers
# dummy values
#nordugrid-queue-gridrunning
$gridrunning = 13;
#nordugrid-queue-gridqueued
$gridqueued = 13;
#LRMSPORT>



#<LRMSPORT
# nordugrid-queue-status
# determine the $lrms_queue_status from the LRMS
# set it to 'active' if the queue can accept jobs
$lrms_queue_status = "active";
#LRMSPORT>


# loop over the mapped local unix users and determine the LRMS-dependent
# freecpus and queuelength for each of them. 
# Use the %users hash which contains the grid user -> local unix user
# mapping info:  $users{Grid_User_SN}=mapped_unix_id
# Take into account queue limits (cpu & time),
# physically available cpus (values determined above) or use LRMS specific method
# Store the result in two hashes, take care of the syntax of the freecpus: cpus[:minutes] 
# here the :minutes part is optional and used for specifying time limit on CPUs.
# $freecpus_for_mapped_users{unix_username}="13:33"
# $user_jobs_queued{$unix_username}=13


foreach $mapped_unixid ( values %users ) {
    # dummy values, these should be calculated for every mapped unix users from the LRMS info
    #nordugrid-authuser-freecpus
    $users_freecpus{$mapped_unixid} = "13:133";
    #nordugrid-authuser-queuelength	    
    $users_jobs_queued{$mapped_unixid} = 13;
}
    
#LRMSPORT>

&queueldif;


&jobsldif;


&usersldif;



my $runtime =  time - $^T;
if ($config{loglevel} == 2) {
   &infosys_shared::write_log("execution time: $runtime");
}
elsif ($config{loglevel} = 1 and $runtime >= 4 ) {  
   &infosys_shared::write_log("SLOW script: $runtime");  
}
