"""
GIIS module. Contains various stuff to use GIIS servers.
"""

import arcbindings as arc
import logger
import errors


class GIISError(errors.ARCError):
	"""Raised when an error concerning GIIS happens."""
	pass


def get_giis_list(giis_url_list = None):
	"""
	Retrieves and returns a list of giis servers. The optional
	giis_url_list specify one or more giis servers. If given
	only these servers will be returned.
	"""
	log = logger.Logger()

	giis_url_vector = arc.string_vector()
	if giis_url_list is not None:
		log.debug("Recieved a giis list - %s" % str(giis_url_list))
		for giis_url in giis_url_list:
			giis_url_vector.append(giis_url)

	giis_vector = arc.giis_vector()
	log.debug("Fetching GIIS list")
	error_code = arc.GetGiis(giis_url_vector, giis_vector)
	if error_code:
		raise GIISError("Error retrieving GIIS list")

	return giis_vector
