"""
This module contains the user proxy representations.
"""

import os

import arcbindings as arc
import errors
import util


def get_user_proxy_path():
	"""
	Returns the default path of user proxy certificate.
	FIXME should read some enviroment variable to see if it exists,
	since that should be used if set.
	"""
	_PROXY_PATH_PREFIX = "/tmp/x509up_u" # FIXME will not work on win
	uid = os.getuid()
	return _PROXY_PATH_PREFIX + str(uid)


class UserProxyError(errors.ARCError):
	"""
	Raised when an error occurs with the user proxy certificate.
	"""


class UserProxy(arc.CertInfo):
	"""
	Represents a user proxy certificate. Basically a proxy for the C++
	class in the ARC.
	"""
	__metaclass__ = util.MetaSingleton # One is enough
	
	def __init__(self, proxy_url=None):
		object.__init__(self)
		# Setup CertInfo class
		if proxy_url is None:
			arc.CertInfo.__init__(self)
		else:
			arc.CertInfo.__init__(self, proxy_url)
		# Check for expiration
		if self.expired():
			raise UserProxyError("User proxy certificate expired or invalid")

