"""
This file contains various utility functions and such, that does not
belong any particular place.
"""


import errors


def get_cluster_name(jobid):
	"""
	Given a jobid this function returns a string containing the
	name of cluster embedded in the in url. For instance if you have:
	gsiftp://benedict.aau.dk:2811/jobs/224771082306374719659686
	This function returns 'benedict.aau.dk'
	"""
	_JOBID_PREFIX = "gsiftp://"
	if not jobid.startswith(_JOBID_PREFIX):
		raise errors.JobidError("Invalid jobid (%s)" % jobid)

	# If we have a colon after the prefix we use that for delimiter
	colon_index = jobid.find(":", len(_JOBID_PREFIX) + 1)
	if colon_index != -1: # See if we found something
		return jobid[len(_JOBID_PREFIX):colon_index]

	# Else we go looking for a slash for delimiter
	slash_index = jobid.find("/", len(_JOBID_PREFIX) + 1)
	if slash_index == -1:
		raise errors.JobidError("Invalid jobid (%s)" % jobid)

	return jobid[len(_JOBID_PREFIX):slash_index]


def get_short_jobid(jobid):
	"""
	Given a job id, this functions returns the short job id,
	i.e., the job id without cluster path, just the number.
	"""
	last_slash = jobid.rfind("/")
	if last_slash == -1:
		raise errors.JobidError("Invalid jobid (%s)" % jobid)

	return jobid[last_slash+1:]


class MetaSingleton(type):
	"""
	Singleton Metaclass. You should use this if you need to create a
	singleton. Especially if it should be a thread, since this is the
	only way that I (htj) could figure out to create one. Tried the
	function trick and mixin, this one works!

	Also "metaclass" sounds  way to cool to not be used :-)
	To use:
	class MyClass:
		__metaclass__ = MetaSingleton
		# Do whatever you normally do
	"""
	def __init__(cls, name, bases, dict):
		super(MetaSingleton, cls).__init__(name, bases, dict)
		cls.instance = None

	def __call__(cls, *args, **kw):
		if cls.instance is None:
			cls.instance = super(MetaSingleton, cls).__call__(*args, **kw)
		return cls.instance 

