#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <iostream>
#include <vector>
#include <string>
#include <fstream>

#include "Giis.h"
#include "LdapQuery.h"
#include "Preferences.h"


Giis::Giis (const string & url) {

  if (url.substr (0, 7) != "ldap://") return;
  int pos0 = url.find_first_of (":/", 7);
  if (pos0 == string::npos) return;
  host = url.substr (7, pos0 - 7);
  int pos = url.find ('/', pos0);
  if (pos == pos0)
    port = 2135;
  else
    port = atoi (url.substr (pos0 + 1, pos - pos0 - 1).c_str());
  while (pos != string::npos) {
    int nextpos = url.find ('/', pos + 1);
    int apos = (nextpos == string::npos) ? url.length() : nextpos;
    if (basedn.empty())
      basedn = url.substr (pos + 1, apos - pos - 1);
    else
      basedn = url.substr (pos + 1, apos - pos - 1) + ',' + basedn;
    pos = nextpos;
  }
}


Giis::Giis (const string & host_,
	    int port_,
	    const string & basedn_) : host (host_),
				      port (port_),
				      basedn (basedn_) {}


bool Giis::operator== (const Giis & giis) const {

  return ((host == giis.host) &&
	  (port == giis.port) &&
	  (basedn == giis.basedn));
}


bool Giis::operator! () const { return host.empty() || basedn.empty(); }


int Giis::Connect (const string & usersn, bool anonymous,
		      int timeout, int debug) {

  return ldap.Connect (host, port, usersn, anonymous, timeout, debug);
}


int Giis::Query (const string & usersn, int timeout, int debug) {

  const string filter;
  const string attr = "giisregistrationstatus";
  vector <string> attrs;
  attrs.push_back (attr);

  return ldap.Query (basedn, filter, attrs, LdapQuery::base, timeout, debug);
}


int Giis::Result (void callback (const string & attr,
				 const string & value,
				 void * ref),
		  void * ref, int timeout, int debug) {

  return ldap.Result (callback, ref, timeout, debug);
}


int GetGiises (const vector <string> & giisurls, vector <Giis> & giislist) {

  if (!giisurls.empty()) {
    for (vector <string>::const_iterator vsi = giisurls.begin();
	 vsi != giisurls.end(); vsi++) {
      Giis giis (*vsi);
      if (!giis) {
	cerr << "Error: Syntax error in giisurl: " << *vsi << endl;
	return 1;
      }
      else
	giislist.push_back (giis);
    }
    return 0;
  }

  int i = 0;
  bool found = false;
  while (!found) {
    string filename;
    switch (i) {
    case 0:
      filename = GetEnv ("HOME");
      filename.append ("/.nggiislist");
      break;
    case 1:
      filename = GetEnv ("NORDUGRID_LOCATION");
      if (filename.empty()) filename = DEFAULT_NORDUGRID_LOCATION;
      filename.append ("/etc/giislist");
      break;
    case 2:
      filename = "/etc/giislist";
      break;
    case 3:
      cerr << "Error: No giislist found at standard locations" << endl;
      return 1;
      break;
    }
    i++;
    ifstream giislistfile (filename.c_str());
    if (!giislistfile) continue;
    string line;
    while (getline (giislistfile, line)) {
      if (line.empty()) continue;
      if (line[0] == '#') continue;
      Giis giis (line);
      if (!giis) {
	cerr << "Error: Syntax error in " << filename << ": " << line << endl;
	return 1;
      }
      else
	giislist.push_back (giis);
    }
    giislistfile.close();
    if (giislist.empty()) {
      cerr << "Error: Giislist file " << filename << " contains no giises"
	   << endl;
      return 1;
    }
    return 0;
  }
}
