#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <iostream>
#include <string>
#include <vector>

#include "Preferences.h"
#include "CertInfo.h"
#include "DateTime.h"
#include "MdsQuery.h"

#define __UI_LIBRARY__
#include "ui_uploader.h"

#include "ngui.h"


extern "C"
int ngcleanxx (const vector <string> & jobs,
	       const vector <string> & joblists,
	       const vector <string> & clusterselect_,
	       const vector <string> & clusterreject_,
	       const vector <string> & status,
	       const bool all,
	       const bool force,
	       int timeout,
	       int debug,
	       const bool anonymous) {

  vector <string> clusterselect = ResolveAliases (clusterselect_);
  vector <string> clusterreject = ResolveAliases (clusterreject_);

  vector <string> jobids;
  vector <Cluster> clusterlist;

  int error = 0;
  ActivateGlobus();

  // setting defaults

  if (timeout == UNDEFINED) timeout = iGetEnv ("NGTIMEOUT");
  if (timeout == UNDEFINED) timeout = iGetDef ("NGTIMEOUT");
  if (timeout == UNDEFINED) timeout = DEFAULT_TIMEOUT;

  if (debug == UNDEFINED) debug = iGetEnv ("NGDEBUG");
  if (debug == UNDEFINED) debug = iGetDef ("NGDEBUG");
  if (debug == UNDEFINED) debug = 0;

  // get user info

  CertInfo user;
  if (!user) {
    error = 1;
    goto errorexit;
  }

  if (debug) {
    cout << "User subject name: " << user.GetSN() << endl;
    cout << "Remaining proxy lifetime: "
	 << Period (user.TimeLeft()) << endl;
  }

  // finding the information

  jobids = GetJobIDs (jobs, joblists, clusterselect, clusterreject, status,
		      all);

  if (all && jobids.empty()) {
    cout << "ngclean: No jobs" << endl;
    goto errorexit;
  }

  if (jobids.empty()) {
    cerr << "ngclean: no valid jobids given" << endl;
    cerr << "Use \"ngclean -help\" for help" << endl;
    error = 1;
    goto errorexit;
  }

  for (vector <string>::iterator vsi = jobids.begin();
       vsi != jobids.end(); vsi++) {
    int pos0 = 0;
    if (vsi->substr (0, 9) == "gsiftp://") pos0 = 9;
    int pos = vsi->find_first_of (":/", pos0);
    if (pos == string::npos) {
      cerr << "ngclean: invalid jobid: " << *vsi << endl;
      error = 1;
      continue;
    }
    string clustername = vsi->substr (pos0, pos - pos0);
    bool found = false;
    for (vector <Cluster>::iterator cli = clusterlist.begin();
	 !found && cli != clusterlist.end(); cli++)
      if (clustername == cli->GetName()) found = true;
    if (!found) clusterlist.push_back (clustername);
  }

  if (clusterlist.empty()) {
    cerr << "ngclean: no valid jobids given" << endl;
    error = 1;
    goto errorexit;
  }

  FindClusterInfo (clusterlist, Mds::JobManipulation, user.GetSNx(), anonymous,
		   timeout, debug);

  for (vector <string>::iterator vsi = jobids.begin();
       vsi != jobids.end(); vsi++) {

    Cluster * c = NULL;
    string jobstatus;

    bool found = false;
    bool selected = false;

    for (vector <Cluster>::iterator cli = clusterlist.begin();
	 !found && cli != clusterlist.end(); cli++)
      for (vector <Queue>::iterator qli = cli->queues.begin();
	   !found && qli != cli->queues.end(); qli++)
	for (vector <Job>::iterator jli = qli->jobs.begin();
	     !found && jli != qli->jobs.end(); jli++)
	  if (*vsi == jli->GetId()) {
	    if (jli->IsStatus (status)) {
	      selected = true;
	      c = &*cli;
	      jobstatus = jli->GetStatus();
	    }
	    found = true;
	  }

    if (force && status.empty() && !found) {
      cout << "ngclean: deleting job " << *vsi << endl;
      RemoveJobID (*vsi);
      continue;
    }
    if (!found) {
      cerr << "ngclean: no job with jobid " << *vsi << " found" << endl;
      error = 1;
      continue;
    }
    if (!selected)
      continue;
    if (jobstatus != "FINISHED" && jobstatus != "FAILED" &&
	jobstatus != "CANCELLED" && jobstatus != "DELETED") {
      cerr << "ngclean: job " << *vsi << " has not finished yet." << endl;
      error = 1;
      continue;
    }
    int pos = vsi->rfind ('/');
    if (pos == string::npos) {
      cerr << "ngclean: invalid jobid: " << *vsi << endl;
      error = 1;
      continue;
    }
    string shortid = vsi->substr (pos + 1);

    const vector <string> filenames;

    cout << "ngclean: deleting job " << *vsi << endl;

    char * shortjobid = strdup (shortid.c_str());
    int err = ui_uploader (c->GetContact().c_str(), NULL, &shortjobid, NULL,
			   RSL_ACTION_CLEAN, filenames, debug, timeout);
    free (shortjobid);

    if (err) {
      cerr << "ngclean: job " << *vsi << " could not be deleted." << endl;
      error = 1;
      continue;
    }

    RemoveJobID (*vsi);
  }

 errorexit:
  DeactivateGlobus();
  return error;
}
