/** ngjmsub.cpp - Submiter for the NG Job Manager
 *
 * Copyright 2004 Henrik Thostrup Jensen and Jesper Ryge Leth
 * All rights reserved.
 *
 * This file is part of NGProxy.
 * 
 * NGProxy is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * NGProxy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with NGProxy; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/**
 * NGProxy will change name to "Nunabungo Gravdeha Jelahero Mongyput" /
 * NG Job Manager, or ngjm, but for now it still called NGProxy
 */


#include <iostream>

#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/sendfile.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/un.h>

using namespace std;

int main(int argc, char** argv) {

	// Checj for correct argument
	if (argc != 3) {
		cout << "Usage: " << argv[0] << " file uds_socket\n";
		return 1;
	}

	// Create socket
	int client_socket = socket(AF_UNIX, SOCK_STREAM, 0);
	if (client_socket == -1) {
		cout << "Error creating socket" << endl;
		return 1;
	}

	// Create socket adress struct for unix domain sockets
	struct sockaddr_un address;
	address.sun_family = AF_UNIX;
	strcpy(address.sun_path,argv[2]);

	// Open file
	int file_handle = open(argv[1],O_RDONLY);
	if (file_handle == -1) {
		cout << "Error opening file: " << argv[1] << endl;
		return 1;
	}

	// Connect to file
	int addr_length = strlen(address.sun_path) + sizeof(address.sun_family);
	int retval = connect(client_socket, (struct sockaddr*) &address, addr_length);
	if (retval == -1) {
		cout << "Error connecting" << endl;
		close(file_handle);
		return 1;
	}

	// Get file stats
	struct stat file_stat;
	fstat(file_handle,&file_stat);

	int result = 0;

	// Copy file
	int bytes_copied = sendfile(client_socket, file_handle, (off_t) 0, file_stat.st_size);
	if (bytes_copied == -1) {
		cout << strerror(errno) << endl;
		result = 1;
	}
	else if (bytes_copied != file_stat.st_size) {
		cout << "Incomplete copy - copied " << bytes_copied << " bytes of " << file_stat.st_size << endl;
		result = 2;
	}
	else {
		cout << "File copied to socket" << endl;
	}

	close(client_socket);
	close(file_handle);

	return result;
}
